#!/bin/bash

# This script is called from Java based Job Server
# Job server calls this script with some important information like (IPO)
# Earlier we relied on this IPO information provided over the command line to this script.
# But presence of characters like single quote in user program names caused trouble.
# We are now shifting to env variables based solution, where java based program has already set our required variables.
# Still this script receives ipo on commandline and we haven't phased it out yet.
# This program prepares necessary Fifo etc and calls user program

set -x
set -o pipefail

start=$(date +%s)





# Some helpful debugging info
echo "Current program full path: $(pwd)/$0"
echo "Host name is: $(hostname)"
echo "Current user is: $(whoami)"
echo "No of command line args: $#"
echo "Printing command line args: $*"



inargs=()
pargs=()
outargs=()


# Lets do some sanity checking about env variables required
function envSanityChecking {

    if [[ -z "${__plumb_no_of_inputs}" ]]
    then
        echo "__plumb_no_of_inputs not set."
    fi
    for i in $(seq 1 "${__plumb_no_of_inputs}");
    do
        x="__plumb_i$i"
        if [[ -z "${!x}" ]];
         then echo "$x not set."
              exit 1
        fi
        inargs+=("${!x}")

    done;



    if [[ -z "${__plumb_no_of_outputs}" ]]
    then
        echo "__plumb_no_of_outputs not set."
    fi
    for i in $(seq 1 "${__plumb_no_of_outputs}");
    do
        x="__plumb_o$i"
        if [[ -z "${!x}" ]];
         then echo "$x not set."
              exit 2
        fi
        outargs+=("${!x}")

    done;


    if [[ -z "${__plumb_prog_name}" ]]
    then
        echo "__plumb_prog_name not set."
        exit 3
    fi

    if [[ -z "${__plumb_prog_args}" ]]
    then
        echo "__plumb_prog_args not set."
    fi

}



function checkBinaryAvailability {


        programPath="${__plumb_prog_name}"
        programName=$(basename "$programPath")
        restOfCmd="${__plumb_prog_args}"
        newCmd="./$programName  $restOfCmd"

        /usr/local/bin/hdfs -copyToLocal "$programPath"  ./
        retCode=$?
        if [ $retCode == 0 ]
        then
                chmod +x ./"$programName"
                return $retCode
        fi


        echo "Copying binary from HDFS to local dir for file: $programPath failed with error code $retCode"

        if [ -e "$programPath" ]
        then
                ln -s -f "$programPath"  "$programName"
                return $?
        fi

        echo "$programPath not found on this server. Exiting."
        exit 4

}


function LinkRegExScript {
    if [[ -z "${pathToRegEx}" ]]  #  YARNApplicationMaster should have set it up
    then
        echo "Env variable: pathToRegEx not set."
        return 1
    fi


    ln -s "${pathToRegEx}"  ./regex.py

}



eval envSanityChecking
eval checkBinaryAvailability
# above func sets newCmd
pargs+=("$newCmd")

eval LinkRegExScript



# Some more debugging information
echo -n "Input: "
echo "${inargs[*]}"

echo -n "Program: "
echo "${pargs[*]}"

echo -n "Output: "
echo "${outargs[*]}"



inFifoList=()
outFifoList=()


function finish {
  echo " Cleaning up Fifos... "
  for inF in ${inFifoList[*]}
  do
    [ -e "$inF" ] && rm -f "$inF"
  done

  for outF in ${outFifoList[*]}
  do
    [ -e "$outF" ] && rm -f "$outF"
  done



  date_epochs=$(date +%s)
  duration=$((date_epochs-start))
  echo "duration = $duration"


}

# Ask shell to run finish function on program exit.
trap finish EXIT

##############################################################
# $1 = hdfs file to read from
# $2 = Fifo to write to

function writerChild {

    echo "WriterChild: $1 => $2 "
    /usr/local/bin/hdfs -cat "$1" > "$2"

    writeChildRetCode=$?
    if [ $writeChildRetCode != 0 ]
    then
    	echo "[writerChild] got non zero ret code.  (ret code = $writeChildRetCode)"
    	exit 3
    fi


}
#############
# $1 = Fifo to read from
# $2 = hdfs file to write to

function readerChild {

    echo "ReaderChild: $1 => $2 "
    /usr/local/bin/hdfs -Ddfs.replication=3 -copyFromLocal -f "$1" "$2"
    readChildRetCode=$?



    if [ $readChildRetCode != 0 ]
    then
        echo "[readerChild] got non zero ret code.  (ret code = $readChildRetCode)"
        exit 4
    fi



}
#################################################################
function die {
    echo "$1"
    echo "Existing ..."
    exit 1

}

#################################################################
# Some exports of env variables
PLUMB_INPUTS=$(IFS=,; echo "${inargs[*]}")
export PLUMB_INPUTS
# following for case-insensitive search
shopt -s nocasematch
if [[ " ${outargs[*]} " =~ " None " ]]; then
    export PLUMB_OUTPUTS
    atLeastOneOutputIsNone=true

else
    PLUMB_OUTPUTS=$(IFS=,; echo "${outargs[*]}")
    export PLUMB_OUTPUTS
    atLeastOneOutputIsNone=false
fi
shopt -u nocasematch

printenv



# User program command line
uCmd="${pargs[*]}"


# If just one in and one out, do the simple case


count=$((${#outargs[@]} + ${#inargs[@]}))
echo $count

# following one line done only to facilitate my end-to-end experiment
inF=${inargs[0]}

retCode=0

if [ $count == 2 ]
then
    # To Fix: current using pargs[*].  Need to fix it when having spaces.

    # Hack: Few apps (like rssac wrapper) needs site name and monotonically increasing
    # serial number.  They rely on doing e.g "ls *.message.fsdb" etc.  Fifos with those
    # names will be present for other case, but for this single input, single output case
    # we need to make zero sized files
    # Note added on 2020-05-16: After getting a proper regex.py, and related API
    # (see https://wiki.isi.edu/div7/index.php/User:Aqadeer/windowing/#File_Name_Conventions)
    # user code can now more conveniently get meta data, or site name (that will be now sequence
    # no instance).  Nevertheless I am leaving following three lines intact for backwards compatibility.
    a=$(basename "${inargs[0]}")
    b=$(basename "${outargs[0]}")

    touch ./"$a"  ./"$b"

    ls -lah


    if [ "$atLeastOneOutputIsNone" = true ] ; then
    /usr/local/bin/hdfs -cat "${inargs[0]}" | eval "${pargs[*]}"
    else
        /usr/local/bin/hdfs -cat "${inargs[0]}" | eval "${pargs[*]}" | /usr/local/bin/hdfs -Ddfs.replication=3 -copyFromLocal -f -  "${outargs[0]}"
    fi


    # ToDo: HEADS UP: If user code does not read off input, then due to pipe fail -o, the first part of above pipeline will
    # fail.  We need to think a viable solution to this problem.
    retCode=$((retCode+$?))



    rm -f   ./"$a"   ./"$b"

    exit $retCode

fi



# Make required input Fifo

for inF in "${inargs[@]}";
do
    originalInF=$inF
    inF=$(basename "$inF")
    echo "$inF"
    mkfifo "$inF" || die "Failed to create input Fifo"
    inFifoList+=("$inF")

    writerChild "$originalInF"  "$inF"  &
    children+=($!)

    uCmd="${uCmd} -i./$inF"

done


# Make required output Fifo
# If any of the output queue is None, then no output at all
if [ "$atLeastOneOutputIsNone" = false ] ; then
    for outF in "${outargs[@]}";
    do
        originalOutF=$outF
        outF=$(basename "$outF")
        echo "$outF"
        mkfifo "$outF" || die "Failed to create output Fifo"
        outFifoList+=("$outF")

        readerChild "$outF"  "$originalOutF"  &
        children+=($!)

        uCmd="${uCmd} -o./$outF"

    done
fi

echo "User command line is: $uCmd"

# Run user program
eval "$uCmd"
retCode=$?


# Terminate any stuck child-processes
# That can happen if $uCmd was faulty and child-processing are stuck waiting

if [ $retCode != 0 ]
then
	pkill -P $$
fi

for child in "${children[@]}"
do
    echo "Waiting on child with id: $child"
    wait "$child"

done

exit $retCode
