/*
 * Copyright 2015, Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *    * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *
 *    * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package edu.isi.ant.JobAPI;

import io.grpc.ManagedChannel;
import io.grpc.ManagedChannelBuilder;
import io.grpc.StatusRuntimeException;


import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Scanner;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

import java.io.File;

import io.grpc.netty.GrpcSslContexts;
import io.grpc.netty.NettyChannelBuilder;
import io.netty.handler.ssl.SslContext;
import io.grpc.netty.NegotiationType;

import javax.net.ssl.SSLException;


public class JobAPIClient {
    private static final Logger logger = Logger.getLogger(JobAPIClient.class.getName());

    private final ManagedChannel channel;
    private final JobAPIServiceGrpc.JobAPIServiceBlockingStub blockingStub;

    /**
     * Construct client connecting to HelloWorld server at {@code host:port}.
     */
    public JobAPIClient(String host, int port) throws SSLException {

        String whoami = System.getProperty("user.name");

        String clientKey = String.format("%s/%s/%s",
                                         ServerConfig.getString("clientPrivateKeyDir"),
                                         whoami,
                                         ServerConfig.getString("clientPrivateKeyFileName"));
        if(!ServerConfig.doesLocalFileExists(clientKey))
        {
            System.err.println("Failed to find private key for user: " + whoami + " key looked for = " + clientKey);
            System.exit(1);
        }

        String clientCertificateChain = String.format("%s/%s/%s",
                                                      ServerConfig.getString("clientCertChainDir"),
                                                      whoami,
                                                      ServerConfig.getString("clientCertChainFileName"));
        if(!ServerConfig.doesLocalFileExists(clientCertificateChain))
        {
            System.err.println("Failed to find certificate chain for user: " + whoami + " chain looked for = " + clientCertificateChain);
            System.exit(2);
        }

        File roots_pem = new File(ServerConfig.getString("rootsCertChain"));
        File client_key = new File(clientKey); //new File(ServerConfig.getString("clientPrivateKey"));
        File client_certificate_chain = new File(clientCertificateChain);//new File(ServerConfig.getString("clientCertChain"));

        SslContext sslContext = GrpcSslContexts.forClient()
                .trustManager(roots_pem).keyManager(client_certificate_chain, client_key)
                .build();


        channel = NettyChannelBuilder.forAddress(host, port)
                .negotiationType(NegotiationType.TLS)
                .sslContext(sslContext)
                .build();

        blockingStub = JobAPIServiceGrpc.newBlockingStub(channel);

    }


    /**
     * Construct client for accessing Job API  server using the existing channel.
     */
    JobAPIClient(ManagedChannelBuilder<?> channelBuilder) {
        channel = channelBuilder.build();
        blockingStub = JobAPIServiceGrpc.newBlockingStub(channel);
    }


    public void shutdown() throws InterruptedException {
        channel.shutdown().awaitTermination(5, TimeUnit.SECONDS);
    }


    /**
     * Submit your job to server.
     */
    public void submitJob(String jobFilePath) {
        logger.info("Submitting job: " + jobFilePath + " ...");
        //JobAPIRequest request = JobAPIRequest.newBuilder().setJobHDFSPath(name).build();
        JobAPIRequest request = JobAPIRequest.newBuilder().build();
        JobAPIReply response;
        try {
            response = blockingStub.getHDFSPathForClient(request);
        } catch (StatusRuntimeException e) {
            logger.log(Level.WARNING, "RPC failed: {0}", e.getStatus());
            e.printStackTrace();
            return;
        }
        String message = String.format("result = %s, errCode = %s, message = %s", response.getResult(),
                response.getRetCode(),
                response.getErrorMessage());
        logger.info(message);
        logger.info("Client HDFS path: " + response.getClientHDFSDirPath());


        // Now put user provided YAML job file into the HDFS dir mentioned by server

        if (0 != HDFSHelper.copyFromLocal(jobFilePath, response.getClientHDFSDirPath(), logger)) {
            String errMessage = String.format("Failed to copy %s into HDFS %s.", jobFilePath, response.getClientHDFSDirPath());
            logger.warning(errMessage);
            return;
        }


        // Now do the actual submit call to the server.
        String fileName = jobFilePath.substring(jobFilePath.lastIndexOf("/") + 1);
        String hdfsFileName = String.format("%s/%s", response.getClientHDFSDirPath(), fileName);

        request = JobAPIRequest.newBuilder().setJobHDFSPath(hdfsFileName).build();
        response = blockingStub.submitJob(request);

        message = String.format("result = %s, errCode = %s, message = %s", response.getResult(),
                response.getRetCode(),
                response.getErrorMessage());
        logger.info(message);


    }



    private final String submitJob = "submitJob";
    private final String retractJob = "retractJob";
    private final String listAcceptedJobs = "listAcceptedJobs";
    private final String listSubmittedJobs = "listSubmittedJobs";
    private final String listRejectedJobs = "listRejectedJobs";
    private final String listRetractedJobs = "listRetractedJobs";
    private final String getOptYamlText = "getOptYamlText";
    private final String getOptYamlDiagram = "getOptYamlDiagram";


    private void printUserCommands() {
        System.err.println("Following commands are available.");
        String optYaml = String.format("(This command makes a YAML file in cwd with name: %s)",
                                       ServerConfig.getString("overallOptimizedYAML"));
        String optpng = String.format("(This command makes a PNG file in cwd with name: %s)",
                                      ServerConfig.getString("overallOptimizedDiagram"));
        String commands = "submitJob <fully_qualifed_local_path_to_YAML_job_file> \n" +
                "retractJob <name_of_YAML_job_file_as_returned_by_listAcceptedJobs_fuction> \n" +
                "listAcceptedJobs \n" +
                "listSubmittedJobs \n" +
                "listRejectedJobs \n" +
                "listRetractedJobs \n" +
                "getOptYamlText  "   + optYaml + " \n" +
                "getOptYamlDiagram " + optpng  + " \n";

        System.err.println(commands);


    }

    private boolean parseUserCommand(String tokens[]) {

        boolean result = false;



        if (tokens.length == 0)
        {
            this.printUserCommands();
            return result;
        }

        result = true;

        switch (tokens[0]) {
            case submitJob:
                this.cmdSubmitJob(tokens);
                break;

            case retractJob:
                this.cmdRetractJob(tokens);
                break;

            case listAcceptedJobs:
                this.cmdListAcceptedJobs(tokens);
                break;

            case listSubmittedJobs:
                this.cmdListSubmittedJobs(tokens);
                break;

            case listRejectedJobs:
                this.cmdListRejectedJobs(tokens);
                break;

            case listRetractedJobs:
                this.cmdListRetractedJobs(tokens);
                break;

            case getOptYamlText:
                this.cmdGetOptYamlText(tokens);
                break;

            case getOptYamlDiagram:
                this.cmdGetOptYamlDiagram(tokens);
                break;


            default:
                System.err.println("Unrecognized command.");
                this.printUserCommands();
                result = false;
        }

        return result;

    }




    private void cmdSubmitJob(String tokens[]) {

        if (tokens.length != 2) {
            System.err.println("submitJob only takes one argument: fully qualified local path to YAML job file.");
            System.exit(3);
        }

        if (Files.notExists(Paths.get(tokens[1]))) {
            System.err.println("Not able to locate local file: " + tokens[1]);
            System.exit(4);

        }

        // check syntax on provided yaml file
        if (!MyUtils.processUserJobFile(tokens[1])) {
            System.err.println("Syntax of provided YAML file not correct.");
            System.exit(5);

        }

        // get hdfs path to submit job
        logger.info("Submitting job to server: " + tokens[1] + " ...");
        JobAPIRequest request = JobAPIRequest.newBuilder().build();
        JobAPIReply response;
        try {
            response = blockingStub.getHDFSPathForClient(request);
            String message = String.format("result = %s, errCode = %s, message = %s", response.getResult(),
                    response.getRetCode(),
                    response.getErrorMessage());
            logger.info(message);
            logger.info("Client HDFS path: " + response.getClientHDFSDirPath());

            String submitPath = MyUtils.getClientSubmitDir(response.getClientHDFSDirPath());


            // Now put user provided YAML job file into the HDFS dir mentioned by server
            if (0 != HDFSHelper.copyFromLocal(tokens[1], submitPath, logger)) {
                String errMessage = String.format("Failed to copy %s into HDFS %s.", tokens[1], response.getClientHDFSDirPath());
                logger.warning(errMessage);
                System.exit(6);
            }



          /* I think following submitJob now not needed


          // Now do the actual submit call to the server.
          String jobFilePath = tokens[1];
          String fileName = jobFilePath.substring(jobFilePath.lastIndexOf("/")+1);
          String hdfsFileName = String.format("%s/%s" , response.getClientHDFSDirPath(), fileName);

          request = JobAPIRequest.newBuilder().setJobHDFSPath(hdfsFileName).build();
          response = blockingStub.submitJob(request);

          message = String.format("result = %s, errCode = %s, message = %s", response.getResult(),
                  response.getRetCode(),
                  response.getErrorMessage());
          logger.info(message);

          */


            // indicate to the server that it should fetch new user job(s)
            response = blockingStub.incorporateNewUserJobs(request);
            message = String.format("result = %s, errCode = %s, message = %s", response.getResult(),
                    response.getRetCode(),
                    response.getErrorMessage());
            logger.info(message);
            logger.info("Server informed about new job.  " +
                    "Server will incorporate new job(s) when currently running tasks on cluster complete.");
        } catch (StatusRuntimeException e) {
            logger.log(Level.WARNING, "RPC failed: {0}", e.getStatus());
            e.printStackTrace();
            System.exit(7);


        }


    }


    private void cmdRetractJob(String tokens[]) {

        if (tokens.length != 2) {
            System.err.println("retractJob only takes one argument: Name of job file (as provided by list function).");
            System.exit(8);
        }

        String jobName = MyUtils.getBaseName(tokens[1]);


        // get hdfs base dir for this user
        JobAPIRequest request = JobAPIRequest.newBuilder().build();
        JobAPIReply response;
        try {
            response = blockingStub.getHDFSPathForClient(request);
            String message = String.format("result = %s, errCode = %s, message = %s", response.getResult(),
                    response.getRetCode(),
                    response.getErrorMessage());
            logger.info(message);
            logger.info("Client HDFS path: " + response.getClientHDFSDirPath());


            String acceptPath = String.format("%s/%s", MyUtils.getClientAcceptDir(response.getClientHDFSDirPath()),
                    jobName);
            String retractPath = MyUtils.getClientRetractDir(response.getClientHDFSDirPath());

            if(!ServerConfig.doesHDFSFileExists(acceptPath))
            {
                System.err.println("Job not found with name:" + acceptPath);
                return;
            }

            HDFSHelper.mvOneFile(acceptPath, retractPath, logger);


            // indicate to the server that it should retract job
            response = blockingStub.incorporateNewUserJobs(request);
            message = String.format("result = %s, errCode = %s, message = %s", response.getResult(),
                    response.getRetCode(),
                    response.getErrorMessage());
            logger.info(message);
            logger.info("Server informed about job retraction.  " +
                    "Server will incorporate changes when currently running tasks on cluster complete.");
        } catch (StatusRuntimeException e) {
            logger.log(Level.WARNING, "RPC failed: {0}", e.getStatus());
            e.printStackTrace();
            System.exit(9);


        }


    }


    private void cmdListAcceptedJobs(String tokens[]) {

        if (tokens.length != 1) {
            System.err.println("No args allowed for this command.");
            System.exit(10);
        }


        // get hdfs base dir for this user
        JobAPIRequest request = JobAPIRequest.newBuilder().build();
        JobAPIReply response;
        try {
            response = blockingStub.getHDFSPathForClient(request);
            String message = String.format("result = %s, errCode = %s, message = %s", response.getResult(),
                    response.getRetCode(),
                    response.getErrorMessage());
            logger.info(message);
            logger.info("Client HDFS path: " + response.getClientHDFSDirPath());


            String acceptPath = MyUtils.getClientAcceptDir(response.getClientHDFSDirPath());

            ArrayList<String> jobs = HDFSHelper.listFiles(acceptPath, logger);

            jobs.forEach(System.out::println);

            ;
        } catch (StatusRuntimeException e) {
            logger.log(Level.WARNING, "RPC failed: {0}", e.getStatus());
            e.printStackTrace();
            System.exit(11);


        }

    }


    private void cmdListSubmittedJobs(String tokens[]) {

        if (tokens.length != 1) {
            System.err.println("No args allowed for this command.");
            System.exit(12);
        }


        // get hdfs base dir for this user
        JobAPIRequest request = JobAPIRequest.newBuilder().build();
        JobAPIReply response;
        try {
            response = blockingStub.getHDFSPathForClient(request);
            String message = String.format("result = %s, errCode = %s, message = %s", response.getResult(),
                    response.getRetCode(),
                    response.getErrorMessage());
            logger.info(message);
            logger.info("Client HDFS path: " + response.getClientHDFSDirPath());


            String submitPath = MyUtils.getClientSubmitDir(response.getClientHDFSDirPath());

            ArrayList<String> jobs = HDFSHelper.listFiles(submitPath, logger);

            jobs.forEach(System.out::println);


        } catch (StatusRuntimeException e) {
            logger.log(Level.WARNING, "RPC failed: {0}", e.getStatus());
            e.printStackTrace();
            System.exit(13);


        }

    }


    private void cmdListRejectedJobs(String tokens[]) {

        if (tokens.length != 1) {
            System.err.println("No args allowed for this command.");
            System.exit(14);
        }


        // get hdfs base dir for this user
        JobAPIRequest request = JobAPIRequest.newBuilder().build();
        JobAPIReply response;
        try {
            response = blockingStub.getHDFSPathForClient(request);
            String message = String.format("result = %s, errCode = %s, message = %s", response.getResult(),
                    response.getRetCode(),
                    response.getErrorMessage());
            logger.info(message);
            logger.info("Client HDFS path: " + response.getClientHDFSDirPath());


            String rejectPath = MyUtils.getClientRejectDir(response.getClientHDFSDirPath());

            ArrayList<String> jobs = HDFSHelper.listFiles(rejectPath, logger);

            jobs.forEach(System.out::println);

            ;
        } catch (StatusRuntimeException e) {
            logger.log(Level.WARNING, "RPC failed: {0}", e.getStatus());
            e.printStackTrace();
            System.exit(15);


        }

    }

    private void cmdListRetractedJobs(String tokens[]) {

        if (tokens.length != 1) {
            System.err.println("No args allowed for this command.");
            System.exit(16);
        }


        // get hdfs base dir for this user
        JobAPIRequest request = JobAPIRequest.newBuilder().build();
        JobAPIReply response;
        try {
            response = blockingStub.getHDFSPathForClient(request);
            String message = String.format("result = %s, errCode = %s, message = %s", response.getResult(),
                    response.getRetCode(),
                    response.getErrorMessage());
            logger.info(message);
            logger.info("Client HDFS path: " + response.getClientHDFSDirPath());


            String retractPath = MyUtils.getClientRetractDir(response.getClientHDFSDirPath());

            ArrayList<String> jobs = HDFSHelper.listFiles(retractPath, logger);

            jobs.forEach(System.out::println);

            ;
        } catch (StatusRuntimeException e) {
            logger.log(Level.WARNING, "RPC failed: {0}", e.getStatus());
            e.printStackTrace();
            System.exit(17);


        }

    }


    private void cmdGetOptYamlText(String tokens[]) {

        if (tokens.length != 1) {
            System.err.println("No args allowed for this command.");
            System.exit(18);
        }

        String src = ServerConfig.getString("optYamlTextPath");
        //String dst = String.format("/tmp/%s/", System.getProperty("user.name"));
        //ServerConfig.makeLocalDirIfDoesntExist(dst);
        String dst = getClientWorkingDir();


        int retCode = HDFSHelper.copyToLocalWithLocalReplacement(src, dst, logger);

        if (retCode == 0) {
            System.out.println("Copied in local dir a file with name: " + MyUtils.getBaseName(src));
        } else {
            System.err.println("Failed to get optimized Yaml.");

        }


    }


    private void cmdGetOptYamlDiagram(String tokens[]) {

        if (tokens.length != 1) {
            System.err.println("No args allowed for this command.");
            System.exit(19);
        }

        String src = ServerConfig.getString("optYamlDiagramPath");
        //String dst = String.format("/tmp/%s/", System.getProperty("user.name"));
        //ServerConfig.makeLocalDirIfDoesntExist(dst);
        String dst = getClientWorkingDir();


        int retCode = HDFSHelper.copyToLocalWithLocalReplacement(src, dst, logger);

        if (retCode == 0) {
            System.out.println("Copied in local dir a file with name: " + MyUtils.getBaseName(src));
        } else {
            System.err.println("Failed to get optimized Yaml diagram.");

        }


    }


    private String getClientWorkingDir()
    {
        //If client was called via wrapper ANT_GIT/CODE/BigDataProcessing/JobServer/ApplicationServer/my_scripts/job-api-client
        // then originalWorkingDir should be set
        // Else we fall back to usual case.

        String cwd = System.getenv().get("originalWorkingDir");
        if(cwd == null || cwd.isEmpty())
        {
            cwd = "./";
        }
        return cwd;
    }


    @Override
    public void finalize() {
        try {
            this.shutdown();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
    }


    public static void main(String[] args) {

        boolean result = false;


       try {
           Logger log = Logger.getLogger("edu.isi.ant.JobAPI");
           log.setLevel(java.util.logging.Level.FINEST);
           JobAPIClient client = new JobAPIClient(ServerConfig.getString("ApplicationServerHostName"),
                   ServerConfig.getInt("ApplicationServerListeningPort"));


           result = client.parseUserCommand(args);
       }catch (Exception exp)
       {
           exp.printStackTrace();
           System.exit(20);

       }

        System.exit((result == true)?0:1);



    }


}
