/*
 * Copyright 2015, Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *    * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *
 *    * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package edu.isi.ant.JobAPI;


import edu.isi.ant.yamlseranddeser.InputQueueParsingFailedException;
import edu.isi.ant.yamlseranddeser.Window;
import edu.isi.ant.yamlseranddeser.YAMLSerAndDeSer;
import io.grpc.Server;
import io.grpc.stub.StreamObserver;


import java.io.IOException;
import java.io.File;
import java.util.*;
import java.util.logging.Logger;

import io.grpc.netty.GrpcSslContexts;
import io.grpc.netty.NettyServerBuilder;
import io.netty.handler.ssl.ClientAuth;
import io.grpc.*;
import org.apache.commons.io.FileUtils;
import org.apache.commons.configuration2.Configuration;

import javax.naming.InvalidNameException;
import javax.naming.ldap.LdapName;
import javax.naming.ldap.Rdn;
import javax.net.ssl.*;

import edu.isi.ant.yamlseranddeser.WindowParsingFailedException;

/**
 * Job API Implementation
 */
public class JobAPIService {
    private static final Logger logger = Logger.getLogger(JobAPIService.class.getName());
    private static final Configuration CONF = ServerConfig.getConfiguration();
    // Following key declared here because reference equality is used by context.
    private static final Context.Key<SSLSession> SSL_SESSION_CONTEXT = Context.key("SSLSession");
    private static final String HDFSClientBaseDirPath = CONF.getString("HDFSClientBaseDirPath");
    private static final String HDFSOptBaseDirPath = CONF.getString("HDFSOptBaseDirPath");
    private Server server;

    private static YAMLSerAndDeSer yamlProcessor = new YAMLSerAndDeSer();
    private static QueueHelper qHelper = new QueueHelper();
    private static LumpinessAwareFastDispatcher dispatcher = null;
    private static Thread thread = null;

    private static String globalOptYAML = setGlobalOptYAML();

    private static Map<String, edu.isi.ant.yamlseranddeser.Window> allWindows = null;
    private static WindowClient windowClient = null;

    private static String setGlobalOptYAML() {
        //String workingDir = System.getProperty("user.dir");
        return String.format("%s/%s", CONF.getString("scratchSpace"), CONF.getString("overallOptimizedYAML"));

    }

    /**
     * Main launches the server from the command line.
     */
    public static void main(String[] args) throws IOException, InterruptedException {
        //Do some sanity checking before starting
        ServerConfig.verify(logger);

        //Del any lingering scratch files
        MyUtils.emptyScratch(ServerConfig.getString("scratchSpace"));
        MyUtils.emptyScratch(ServerConfig.getString("scratchSpaceForYAMLs"));

        //Del HDFS scratch
        HDFSHelper.empytDir(ServerConfig.getString("HDFSScratchSpace"));



        final JobAPIService server = new JobAPIService();
        server.initializeServer();
        server.start();
        server.blockUntilShutdown();
    }


    private void initializeServer() {
        //(0) delete pipeline.yaml from cwd, if it exists.
        MyUtils.delFile(this.globalOptYAML);


        //(1) Get list of all newly submitted user jobs
        ArrayList<String> submittedJobs = HDFSHelper.listFiles(CONF.getString("HDFSClientDirSubmittedSearch"), logger);
        submittedJobs.forEach(System.out::println);


        // (2) Fetch yaml jobs one-by-one in a local pipelie.yaml file and try to get them compiled
        //Jobs with same name as in accepted might be present in submitted.  We will give submitted jobs
        //      precedence over accepted ones.


        for (String job : submittedJobs) {
            String l1 = MyUtils.getDirName(job);
            String l2 = MyUtils.getDirName(l1);
            String user = MyUtils.getBaseName(l2);
            boolean result = this.isUserJobAcceptable(job, user);
            this.moveUserJobs(job, result, logger);

        }

        // Jobs that are submitted + have same name as some job in accepted would have replaced it.

        // (3) Get list of all old (but yet not retracted) usr jobs
        ArrayList<String> previousAcceptedJobs = HDFSHelper.listFiles(CONF.getString("HDFSClientDirAcceptedSearch"), logger);
        previousAcceptedJobs.forEach(System.out::println);


        for (String job : previousAcceptedJobs) {
            String l1 = MyUtils.getDirName(job);
            String l2 = MyUtils.getDirName(l1);
            String user = MyUtils.getBaseName(l2);
            boolean result = this.isUserJobAcceptable(job, user);
            // if due to some reason, earlier accepted job got issues, move it to rejected
            this.moveUserJobs(job, result, logger);


        }


        // If nothing in submitted or accepted, make a dummy pipeline.yaml file to avoid errors later on
        if (ServerConfig.doesLocalFileExists(this.globalOptYAML)) {


            // If isolated parsing okay, then add to optimized dag

            if (!MyUtils.processUserJobFile(yamlProcessor, this.globalOptYAML)) {
                String errMessage = String.format("[%s] %s Failed.", Thread.currentThread().getStackTrace()[1],
                        "processUserJobFile");
                System.err.println(errMessage);
                System.exit(2);

            }


            try {
                String scratch = CONF.getString("scratchSpace");
                String officialYAML = String.format("%s/%s", scratch, "pipeline.yaml");
                String png = String.format("%s/%s", scratch, "pipeline.png");

                yamlProcessor.writeYAMLFile(officialYAML);
                String optDot = String.format("%s/%s", scratch, "opt.dot");
                yamlProcessor.makeDotGraph("OptimizedPipeline", optDot, png);

                String optyaml = CONF.getString("optYamlTextPath");
                String optpng = CONF.getString("optYamlDiagramPath");
                HDFSHelper.mvOneFileFromLocal(officialYAML, optyaml, logger);
                HDFSHelper.mvOneFileFromLocal(png, optpng, logger);
            } catch (IOException e) {
                e.printStackTrace();
            }
        }


        // Deactivate all old queues before moving forward.  This step becomes especially necessary when user has
        // retracted something and server is restarting.  We want data flow to stop at db level as well.
        qHelper.deactivateAllQueuesRepeatOnError(logger);


        // Contact Q-Server to make necessary queues
        Map<String, Integer> dupTable = yamlProcessor.getUniqueQueuesInSystem(true);


        for (String q : dupTable.keySet()) {
            int val = dupTable.get(q);
            qHelper.createQueueRepeatOnError(q, logger);
            qHelper.activateQueueRepeatOnError(q, logger);

            if (val > 1) {
                for (int i = 1; i <= val; i++) {
                    String derivedQ = String.format("%s_%s", q, i);
                    qHelper.createQueueRepeatOnError(derivedQ, logger);
                    qHelper.activateQueueRepeatOnError(derivedQ, logger);
                    qHelper.connectDerivedQueueRepeatOnError(q, derivedQ, logger);
                }

            }
        }

        // queue are made at this point.  Following file just a tag to facilitate experiments
                
        String qMade = String.format("%s/%s", CONF.getString("scratchSpace"), "/q_made/done");
        JobAPIService.makeFile(qMade);


        // We do some initial set-up for any windows here.  Remaining work will be in scheduler
        // Get windowing needs from YAML and inform Q server about them
        this.allWindows = yamlProcessor.getWindowsInSystem(true);
        try {
            this.windowClient = new WindowClient();
        } catch (SSLException e) {
            e.printStackTrace();
            logger.info("Failed to create window client. Exiting...");
            System.exit(2);
        }
        if (!this.windowClient.startMaintainingWindowState(this.allWindows))
        {
            logger.info("Failed to contact Q server for window management. Exiting...");
            System.exit(3);

        }


        Map<String, edu.isi.ant.yamlseranddeser.Stage> dispTable = yamlProcessor.getDispatchTable(dupTable, true);

        if (dispatcher == null) {
            dispatcher = new LumpinessAwareFastDispatcher(dupTable, dispTable, allWindows,logger);
            thread = new Thread(dispatcher);
            logger.info("Starting Lumpiness Aware Tez Dispatcher...");
            thread.start();

        }

    }


    private void moveUserJobs(String srcHDFSFile, boolean result, Logger logger) {
        String fName = MyUtils.getBaseName(srcHDFSFile);
        String l1Dir = MyUtils.getDirName(srcHDFSFile);
        String l2Dir = MyUtils.getDirName(l1Dir);
        String userName = MyUtils.getBaseName(l2Dir);

        String acceptedPath = String.format("%s/%s/%s", CONF.getString("HDFSClientBaseDirPath"), userName, CONF.getString("accepted"));
        String rejectedPath = String.format("%s/%s/%s", CONF.getString("HDFSClientBaseDirPath"), userName, CONF.getString("rejected"));

        if (0 != HDFSHelper.doesXExists(acceptedPath, true)) {
            HDFSHelper.makeDir(acceptedPath, userName, logger);
        }
        if (0 != HDFSHelper.doesXExists(rejectedPath, true)) {
            HDFSHelper.makeDir(rejectedPath, userName, logger);
        }

        if (result) {
            HDFSHelper.mvOneFile(srcHDFSFile, acceptedPath, logger);
        } else {
            HDFSHelper.mvOneFile(srcHDFSFile, rejectedPath, logger);
        }

    }


    private void start() throws IOException {
        Logger log = Logger.getLogger("io.grpc");
        log.setLevel(java.util.logging.Level.FINEST);
        // Some exceptions will show up at FINEST level of logging but this level is good for debugging
        //log.setLevel(java.util.logging.Level.INFO);


        /*  History:
        // Earlier hard-link-emulation code was put in hdfs manually
        // and fetched out here.  It was causing lot of delay.  So now
        // someone manually rsyncs code on local servers and in this code
        // we assume it is present there.
        //fetch out Queueing system out of HDFS
        if(!MyUtils.fetchQueuingSystemCode(logger))
        {
          logger.info("Failed to fetch out Queuing system code out of HDFS.");
          return;
        }
        */


        /* The port on which the server should run */
        int port = CONF.getInt("ApplicationServerListeningPort");


        File privateKeyFile = new File(CONF.getString("serverPrivateKey"));
        File certChainFile = new File(CONF.getString("serverCertChain"));
        File clientCertChainFile = new File(CONF.getString("rootsCertChain"));


        server = NettyServerBuilder.forPort(port)
                .sslContext(GrpcSslContexts.forServer(certChainFile, privateKeyFile)
                        .trustManager(clientCertChainFile)
                        .clientAuth(ClientAuth.REQUIRE).build())
                //.addService(new GreeterImpl())
                .addService(ServerInterceptors.intercept(new JobAPIServiceImpl(), new MyInterceptor()))
                .build()
                .start();


        logger.info("Server started, listening on " + port);


        Runtime.getRuntime().addShutdownHook(new Thread() {
            @Override
            public void run() {
                // Use stderr here since the logger may have been reset by its JVM shutdown hook.
                System.err.println("*** shutting down gRPC server since JVM is shutting down");
                JobAPIService.this.stop();
                System.err.println("*** server shut down");
                thread = null;
            }
        });
    }


    private void stop() {
        if (server != null) {
            server.shutdown();
        }
    }


    /**
     * Await termination on the main thread since the grpc library uses daemon threads.
     */
    private void blockUntilShutdown() throws InterruptedException {
        if (server != null) {
            server.awaitTermination();
        }
    }


    public static boolean makeFile(String fullyQualifiedName) {
        File f = new File(fullyQualifiedName);
        try {
            f.delete(); // del if already exists/
            return f.createNewFile();
        } catch (IOException e) {
            return false;
        }
    }


    private boolean isUserJobAcceptable(String hdfsfullPath, String userName) {
        // (1) copy job file from hdfs to local dir

        // (1.1) We need a tmp dir to keep user files temporarily
        String workingDir = CONF.getString("scratchSpace"); //System.getProperty("user.dir");
        int randNo = (int) (Math.random() * Integer.MAX_VALUE);
        String tmpDir_string = String.format("%s/%s/%s/", workingDir, userName, randNo);
        File tmpDir = new File(tmpDir_string);
        if (false == tmpDir.mkdirs()) {
            String errMessage = String.format("[%s] %s failed.", Thread.currentThread().getStackTrace()[1],
                    "tmpDir.mkdirs()");
            logger.warning(errMessage);
            return false;
        }

        //(1.2) Copy back user job YAML file from HDFS to local tmp dir
        int errCode = HDFSHelper.copyToLocal(hdfsfullPath, tmpDir_string, logger);
        if (0 != errCode) {
            String errMessage = String.format("[%s] %s failed. HDFS Path = %s, Server LocalPath = %s", Thread.currentThread().getStackTrace()[1],
                    "HDFSHelper.copyToLocal()", hdfsfullPath, tmpDir_string);
            logger.warning(errMessage);
            return false;

        }


        // (2) check yaml syntax
        String YAMLFile = hdfsfullPath.substring(hdfsfullPath.lastIndexOf("/") + 1);
        YAMLFile = String.format("%s/%s", tmpDir_string, YAMLFile);

        if (!MyUtils.processUserJobFile(YAMLFile)) {
            String errMessage = String.format("[%s] %s Failed.", Thread.currentThread().getStackTrace()[1],
                    "processUserJobFile");
            logger.warning(errMessage);
            return false;
        }

        // (3) make sure specific queues can be created (Make sure error is not due to already existing queues)
        // (4) If we made queues ourselves above, only then del those queues for now (don't del already existing queues)

        //  Step 3 and 4 are ToDo: because (a) ordering matters (e.g. if a user trying to connect to data a and
        //                                     that data is provided by some other user.  It will wrongly fail.  The
        //                                     solution is to have all data from all users and then check
        //                                 (b) currently single user "aqadeer" and single data type "raw" is in use
        //                                     for all users.  It needs to correct first for user data authorization
        //                                     has to make any sense


        // if everything goes well, (a) copy this user job to overall pipeline.yaml
        //                          (b) move user hdfs job from submitted to accepted
        MyUtils.concatenate(YAMLFile, this.globalOptYAML);


        // if there are parsing or data access errors, copy it in an jobfilename.err file for user and put that as well
        // in rejected job dir

        try {
            FileUtils.deleteDirectory(tmpDir);
        } catch (IOException e) {
            e.printStackTrace();
        }
        // following was not working when dir had contnet in it
        //tmpDir.delete();


        return true;

    }


    static class JobAPIServiceImpl extends JobAPIServiceGrpc.JobAPIServiceImplBase {


        //State needed
        // (1) An object of YAMLSerAndDeSer to keep track of current optimized DAG
        // (2) Per user HDFS dir to keep YAML, Dot file and GraphViz PNG file
        // (3) Global HDFS dir to keep optimized DAG
        // (4) Need to keep DAGs persisted form such that if JobServer dies and come back
        //     it can work correctly.  Also it will need to see what to do if there are
        //     more than one Job Servers (e.g. due to network partitioning etc.)

        @Override
        public void submitJob(JobAPIRequest req, StreamObserver<JobAPIReply> responseObserver) {


            JobAPIReply reply = null;
            String tmpDir_string;

            // (1) Make a temporary dir in cwd  e.g. ./calling_user_id/<some_random_no>/
            // (2) Copy user's job YAML file from HDFS to local dir
            // (3) Pass above dir/file_name to YAML parsing program
            //
            //
            // Get client's id from context
            try {
                // (1) We need a tmp dir to keep user files temporarily
                String CN = MyInterceptor.getCN();
                String workingDir = System.getProperty("user.dir");
                int randNo = (int) (Math.random() * Integer.MAX_VALUE);
                tmpDir_string = String.format("%s/%s/%s/", workingDir, CN, randNo);
                File tmpDir = new File(tmpDir_string);
                if (false == tmpDir.mkdirs()) {
                    String errMessage = String.format("[%s] %s failed.", Thread.currentThread().getStackTrace()[1],
                            "tmpDir.mkdirs()");
                    this.prepareReply(responseObserver, false, 1, errMessage);
                    return;
                }

                logger.info(String.format("Made a tmp dir: %s", tmpDir_string));


                // Copy back user job YAML file from HDFS to local tmp dir
                int errCode = HDFSHelper.copyToLocal(req.getJobHDFSPath(), tmpDir_string, logger);
                if (0 != errCode) {
                    String errMessage = String.format("[%s] %s failed. HDFS Path = %s, Server LocalPath = %s", Thread.currentThread().getStackTrace()[1],
                            "HDFSHelper.copyToLocal()", req.getJobHDFSPath(), tmpDir_string);
                    this.prepareReply(responseObserver, false, errCode, errMessage);
                    return;

                }


                String YAMLFile = req.getJobHDFSPath().substring(req.getJobHDFSPath().lastIndexOf("/") + 1);
                YAMLFile = String.format("%s/%s", tmpDir_string, YAMLFile);

                if (!this.processUserJobFile(new YAMLSerAndDeSer(), YAMLFile)) {
                    String errMessage = String.format("[%s] %s Failed.", Thread.currentThread().getStackTrace()[1],
                            "processUserJobFile");
                    this.prepareReply(responseObserver, false, 2, errMessage);
                    return;
                }

                // mv everything from local tmp dir to user HDFS
                HDFSHelper.mvFromLocal(tmpDir_string, HDFSClientBaseDirPath, logger);


                // If isolated parsing okay, then add to optimized dag
                HDFSHelper.copyToLocal(req.getJobHDFSPath(), tmpDir_string, logger);
                HDFSHelper.makeDir(HDFSOptBaseDirPath, logger);

                if (!this.processUserJobFile(yamlProcessor, YAMLFile)) {
                    String errMessage = String.format("[%s] %s Failed.", Thread.currentThread().getStackTrace()[1],
                            "processUserJobFile");
                    this.prepareReply(responseObserver, false, 2, errMessage);
                    return;
                }

                // mv everything from local tmp dir to opt HDFS
                HDFSHelper.mvFromLocal(tmpDir_string, HDFSOptBaseDirPath, logger);


                tmpDir.delete();


            } catch (SSLPeerUnverifiedException exp) {
                String errMessage = String.format("[%s] %s exception.", Thread.currentThread().getStackTrace()[1],
                        "SSLPeerUnverifiedException");
                this.prepareReply(responseObserver, false, 2, errMessage);
                return;
            } catch (InvalidNameException e) {
                String errMessage = String.format("[%s] %s exception.", Thread.currentThread().getStackTrace()[1],
                        "InvalidNameException");
                this.prepareReply(responseObserver, false, 3, errMessage);
                return;

            }


            // Contact Q-Server to make necessary queues
            Map<String, Integer> dupTable = yamlProcessor.getUniqueQueuesInSystem(true);


            for (String q : dupTable.keySet()) {
                int val = dupTable.get(q);
                qHelper.createQueueRepeatOnError(q, logger);
                qHelper.activateQueueRepeatOnError(q, logger);

                if (val > 1) {
                    for (int i = 1; i <= val; i++) {
                        String derivedQ = String.format("%s_%s", q, i);
                        qHelper.createQueueRepeatOnError(derivedQ, logger);
                        qHelper.activateQueueRepeatOnError(derivedQ, logger);
                        qHelper.connectDerivedQueueRepeatOnError(q, derivedQ, logger);
                    }

                }
            }

            // queue are made at this point
            JobAPIService.makeFile("./q_made/done");


            // get dispatch table
            Map<String, edu.isi.ant.yamlseranddeser.Stage> dispTable = yamlProcessor.getDispatchTable(dupTable, true);

            Map<String, Window> allWindows = yamlProcessor.getWindowsInSystem(true);


            if (dispatcher == null) {
                dispatcher = new LumpinessAwareFastDispatcher(dupTable, dispTable, allWindows, logger);
                thread = new Thread(dispatcher);
                logger.info("Starting Lumpiness Aware Tez Dispatcher...");
                thread.start();

            }

            //ToDo: What to do when we get a new submit call?


            this.prepareReply(responseObserver, true, 0, "OK");
        }


        @Override
        public void incorporateNewUserJobs(JobAPIRequest req,
                                           StreamObserver<JobAPIReply> responseObserver) {
            // Signal dispather that it needs to stop so that next incarnation of server
            // could incorporate new user jobs
            if (dispatcher != null) {
                dispatcher.writeStopSignal();
                this.prepareReply(responseObserver, true, 0, "OK");
            } else {
                this.prepareReply(responseObserver, true, 1, "Server is starting.  Retry later.");

            }


        }


        private void prepareReply(StreamObserver<JobAPIReply> responseObserver,
                                  boolean result,
                                  int retCode,
                                  String errorMessage) {
            JobAPIReply reply = JobAPIReply.newBuilder()
                    .setResult(result)
                    .setRetCode(retCode)
                    .setErrorMessage(errorMessage)
                    .build();


            responseObserver.onNext(reply);
            responseObserver.onCompleted();
        }


        // temporary hack.  must go away
        private void prepareReply2(StreamObserver<JobAPIReply> responseObserver,
                                   boolean result,
                                   int retCode,
                                   String errorMessage,
                                   String HDFSClientDirPath
        ) {
            JobAPIReply reply = JobAPIReply.newBuilder()
                    .setResult(result)
                    .setRetCode(retCode)
                    .setErrorMessage(errorMessage)
                    .setClientHDFSDirPath(HDFSClientDirPath)
                    .build();


            responseObserver.onNext(reply);
            responseObserver.onCompleted();
        }


        @Override
        public void getHDFSPathForClient(JobAPIRequest req, StreamObserver<JobAPIReply> responseObserver) {
            String clientDir = "";
            Context.current();


            try {
                String client = MyInterceptor.getCN();
                clientDir = String.format("%s/%s/", HDFSClientBaseDirPath, client);

                int retCode = HDFSHelper.makeDir(clientDir, client, logger);
                // Also make sub-directories
                retCode += HDFSHelper.makeDir(MyUtils.getClientSubmitDir(clientDir), client, logger);
                retCode += HDFSHelper.makeDir(MyUtils.getClientRetractDir(clientDir), client, logger);
                retCode += HDFSHelper.makeDir(MyUtils.getClientAcceptDir(clientDir), client, logger);
                retCode += HDFSHelper.makeDir(MyUtils.getClientRejectDir(clientDir), client, logger);


                if (retCode != 0) {
                    String errMessage = String.format("[%s] %s failed.", Thread.currentThread().getStackTrace()[1],
                            "HDFSHelper.makeDir(clientDir)");
                    this.prepareReply(responseObserver, false, retCode, errMessage);
                    return;
                }


            } catch (Exception e) {
                String errMessage = String.format("[%s] %s failed.", Thread.currentThread().getStackTrace()[1],
                        "HDFSHelper.makeDir(clientDir)");
                this.prepareReply(responseObserver, false, 2, errMessage);
                return;
            }

            this.prepareReply2(responseObserver, true, 0, "OK", clientDir);

        }


        private boolean processUserJobFile(YAMLSerAndDeSer yamlProcessor, String YAMLFile) {

            String YAMLOut = String.format("%s_%s", YAMLFile, "server.yaml");
            String YAMLDot = String.format("%s_%s", YAMLFile, "server.dot");
            String YAMLGraphViz = String.format("%s_%s", YAMLFile, "server.png");

            String message = String.format("out = %s \n dot = %s \n graph = %s \n", YAMLOut, YAMLDot, YAMLGraphViz);
            logger.info(message);

            boolean result = true;

            try {
                result = yamlProcessor.readYAMLFile(YAMLFile);
                if (!result) return result;

                result = yamlProcessor.writeYAMLFile(YAMLOut);
                if (!result) return result;

                result = yamlProcessor.makeDotGraph("Visualize", YAMLDot, YAMLGraphViz);

            } catch (IOException | WindowParsingFailedException | InputQueueParsingFailedException e) {
                result = false;
            }

            return result;
        }

    }


    static class MyInterceptor implements ServerInterceptor {

        //public final static Context.Key<SSLSession> SSL_SESSION_CONTEXT = Context.key("SSLSession");

        @Override
        public <ReqT, RespT> ServerCall.Listener<ReqT> interceptCall(ServerCall<ReqT, RespT> call,
                                                                     Metadata headers,
                                                                     ServerCallHandler<ReqT, RespT> next) {


            SSLSession sslSession = call.getAttributes().get(Grpc.TRANSPORT_ATTR_SSL_SESSION);
            if (sslSession == null) {
                return next.startCall(call, headers);
            }
            return Contexts.interceptCall(
                    Context.current().withValue(SSL_SESSION_CONTEXT, sslSession), call, headers, next);
        }


        private static String getCN() throws SSLPeerUnverifiedException, InvalidNameException {
            SSLSession sslSession = SSL_SESSION_CONTEXT.get(Context.current());

            String dn = sslSession.getPeerPrincipal().getName();
            LdapName ln = new LdapName(dn);
            String CN = "";

            for (Rdn rdn : ln.getRdns()) {
                if (rdn.getType().equalsIgnoreCase("CN")) {
                    CN = rdn.getValue().toString();
                    break;
                }
            }

            return CN;

        }


    }//end of class MyInterceptor


}
