package edu.isi.ant.JobAPI;


import io.grpc.ManagedChannel;
import io.grpc.lander.queue.api.*;
import io.grpc.netty.GrpcSslContexts;
import io.grpc.netty.NegotiationType;
import io.grpc.netty.NettyChannelBuilder;
import io.netty.handler.ssl.SslContext;

import javax.net.ssl.SSLException;
import java.io.File;
import java.net.URL;
import java.net.URLClassLoader;
import java.time.format.DateTimeFormatter;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;
import java.util.Map;

/**
 * Purpose of this class is to get a direct communication channel (via gRPC)
 * to hard-link-emulation serer (that will now be extended to provide some
 * more functionality for enable windowing).
 * That server is written in python, and we will be using those services from
 * here using java.
 */
public class WindowClient {

    private static final Logger logger = Logger.getLogger(JobAPIClient.class.getName());
    private final ManagedChannel channel;
    private final LanderQueueAPIImplementationGrpc.LanderQueueAPIImplementationBlockingStub blockingStub;
    private String whoami = "Admin"; //System.getProperty("user.name");

    public WindowClient()throws SSLException
    {
        this(ServerConfig.getString("QueueServerHostName"), ServerConfig.getInt("QueueServerListeningPort"));
    }


    public WindowClient(String host, int port) throws SSLException {



        String clientKey = String.format("%s/%s", ServerConfig.getString("adminClientPrivateKeyDir"),
                                         ServerConfig.getString("adminClientPrivateKeyFileName"));
        if(!ServerConfig.doesLocalFileExists(clientKey))
        {
            System.err.println(String.format("Failed to find private key for user: %s key looked for: %s",
                                             whoami, clientKey));
            System.exit(1);
        }

        String clientCertificateChain = String.format("%s/%s", ServerConfig.getString("adminClientCertChainDir"),
                ServerConfig.getString("adminClientCertChainFileName"));
        if(!ServerConfig.doesLocalFileExists(clientCertificateChain)) {
            System.err.println(String.format("Failed to find certificate chain for user: %s chain looked for: %s", 
                                             whoami, clientCertificateChain));
            System.exit(2);
        }

        File roots_pem = new File(ServerConfig.getString("rootsCertChain"));
        File client_key = new File(clientKey); //new File(Constants.clientPrivateKey);
        File client_certificate_chain = new File(clientCertificateChain);//new File(Constants.clientCertChain);

        SslContext sslContext = GrpcSslContexts.forClient()
                .trustManager(roots_pem).keyManager(client_certificate_chain, client_key)
                .build();


        channel = NettyChannelBuilder.forAddress(host, port)
                .negotiationType(NegotiationType.TLS)
                .sslContext(sslContext)
                .build();

        blockingStub = LanderQueueAPIImplementationGrpc.newBlockingStub(channel);

    }

    public void shutdown() throws InterruptedException {
        channel.shutdown().awaitTermination(5, TimeUnit.SECONDS);
    }

    @Override
    public void finalize() {
        try {
            this.shutdown();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
    }


    public boolean startMaintainingWindowState(Map<String, edu.isi.ant.yamlseranddeser.Window> allWindows)
    {

        WindowRequest.Builder requestBuilder = WindowRequest.newBuilder().setCallerId(this.whoami);
        allWindows.forEach((key,value) -> requestBuilder.addWindows(mapFromPlumbWindowTogRPCWindow(key, value)));
        WindowRequest request = requestBuilder.build();
        WindowReply response;

        response = this.blockingStub.startMaintainingWindowState(request);
        System.out.println(response.getRetCode());
        System.out.println(response.getErrorMessage());


        return response.getResult();
    }

    // In the following qID is srcQ names like: message.fsdb  and NOT stage node ID like: message.fsdbTorsscint

    private Window mapFromPlumbWindowTogRPCWindow(String qID, edu.isi.ant.yamlseranddeser.Window plumbWindow)
    {
        System.out.println("qID =" + qID);
        System.out.println(plumbWindow);
        //String dateFormat = "yyyy MM dd HH mm zzz";
        String dateFormatForPython = "%Y %m %d %H %M %Z"; // date format as needed by Python based date time on server side
        String dateFormatForJava = "yyyy MM dd HH mm zzz";
        Window.Builder windowBuilder = Window.newBuilder();
        windowBuilder.setQueueID(qID);
        windowBuilder.setSize(plumbWindow.size);
        windowBuilder.setSizeMetric(plumbWindow.sizeMetric.name());
        windowBuilder.setTimeMetric(plumbWindow.timeMetric.name());
        windowBuilder.setStartTime(plumbWindow.startOfWindow.format(DateTimeFormatter.ofPattern(dateFormatForJava)));
        windowBuilder.setEndTime(plumbWindow.endOfWindow.format(DateTimeFormatter.ofPattern(dateFormatForJava)));
        windowBuilder.setTimeFormat(dateFormatForPython);
        windowBuilder.setMaxWait(plumbWindow.maxWait);
        windowBuilder.setWaitMetric(plumbWindow.maxWaitMetric.name());
        windowBuilder.setCutoff(plumbWindow.cutoffDeliveryRequirement.name());
        windowBuilder.setOutOfOrderWinSched(plumbWindow.outOfOrderScheduling.name());
        windowBuilder.setUseQSeqNoInst(plumbWindow.useQueueSequenceNumberInstance.name());
        windowBuilder.setQSeqNoInstList(plumbWindow.queueSequenceNumberInstancesList);


        return windowBuilder.build();

    }

    public WindowReply getWindowsToSchedule(String queueID)
    {
        //ToDo: Dont't forget to check for wm null from window server
        WindowRequest request = WindowRequest.newBuilder().addWindows(Window.newBuilder().setQueueID(queueID))
                .setCallerId(this.whoami).build();
        WindowReply response = null;

        try
        {
            response = this.blockingStub.getWindowsToSchedule(request);
        }catch(Exception e)
        {
            e.printStackTrace();
            return response;
        }

        System.out.println(response.getRetCode());
        System.out.println(response.getErrorMessage());
        System.out.println(response.getErrorType());
        System.out.println(response);

        return response;
    }

    //WindowOperationRequest

    public WindowOperationResponse changeWindowState(String queueID, long windowNo, String site,
                                                     WindowState currentWindowState,
                                                     WindowState nextWindowState, String application_id)
    {
        //ToDo: Dont't forget to check for wm null from window server
        WindowOperationRequest request = WindowOperationRequest.newBuilder().setQueueID(queueID)
                                        .setWindowNo(windowNo)
                                        .setQSeqNoInst(site)
                                        .setCurrentWindowState(currentWindowState)
                                        .setNextWindowState(nextWindowState)
                                        .setCallerId(this.whoami)
                                        .setApplicationId(application_id).build();
        WindowOperationResponse response;

        response = this.blockingStub.changeWindowState(request);

        System.out.println(response.getRetCode());
        System.out.println(response.getErrorMessage());
        System.out.println(response.getErrorType());
        System.out.println(response);

        return response;

    }

    public WindowOperationResponse unreserveWindow(String queueID, long windowNo)
    {
        //ToDo: Dont't forget to check for wm null from window server
        WindowOperationRequest request = WindowOperationRequest.newBuilder().setQueueID(queueID)
                .setWindowNo(windowNo)
                .setCallerId(this.whoami).build();
        WindowOperationResponse response;

        response = this.blockingStub.unreserveWindow(request);

        System.out.println(response.getRetCode());
        System.out.println(response.getErrorMessage());
        System.out.println(response.getErrorType());
        System.out.println(response);

        return response;

    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Test functions

    public void testCommunication()
    {
        logger.info("Testing communication ...");
        //rpc SayTwoWayStreamingHello (stream HelloRequest) returns (stream HelloReply) {}
        HelloRequest request = HelloRequest.newBuilder().setName("Greetings from Plum App Server!").build();
        HelloReply response;

        response = this.blockingStub.sayHello(request);
        System.out.println(response.getMessage());

    }


    private void startMaintainingWindowState()
    {
        WindowRequest request = WindowRequest.newBuilder().addWindows(Window.newBuilder().setQueueID("Test Queue"))
                                .setCallerId(this.whoami).build();
        WindowReply response;

        for(int i = 0; i < 5; i++)
        {
            response = this.blockingStub.startMaintainingWindowState(request);
            System.out.println(response.getRetCode());
            System.out.println(response.getErrorMessage());
        }
    }


    private void getWindowsToSchedule()
    {
        //ToDo: Dont't forget to check for wm null from window server
        WindowRequest request = WindowRequest.newBuilder().addWindows(Window.newBuilder().setQueueID("rsscint"))
                .setCallerId(this.whoami).build();
        WindowReply response;


        response = this.blockingStub.getWindowsToSchedule(request);
        System.out.println(response.getRetCode());
        System.out.println(response.getErrorMessage());
        System.out.println(response.getErrorType());
        System.out.println(response);


    }


    public static void main(String[] args) {

        boolean result = false;


        try {
            Logger log = Logger.getLogger("edu.isi.ant.JobAPI");
            log.setLevel(java.util.logging.Level.FINEST);
            WindowClient client = new WindowClient(ServerConfig.getString("QueueServerHostName"),
                                                   ServerConfig.getInt("QueueServerListeningPort"));


            //result = client.parseUserCommand(args);
            //client.testCommunication();
            //client.startMaintainingWindowState();
            client.getWindowsToSchedule();


        }catch (Exception exp)
        {
            exp.printStackTrace();
            System.exit(20);

        }

        System.exit((result == true)?0:1);



    }


    public static void printClassPath () {

        ClassLoader cl = ClassLoader.getSystemClassLoader();

        URL[] urls = ((URLClassLoader)cl).getURLs();
        int i = 0;

        for(URL url: urls){
            System.out.println(url.getFile());
            i++;
        }
        System.out.println("**** There are " + i + " entries in class path!");

    }
}
