// code copied from: https://github.com/alexholmes/hiped2/blob/master/src/main/java/hip/ch10/dstat/ApplicationMaster.java
// and then modified

package edu.isi.ant.JobAPI;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;
import java.util.logging.SimpleFormatter;
import java.util.logging.StreamHandler;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import edu.isi.ant.yamlseranddeser.Stage;
import io.grpc.lander.queue.api.RegExReply;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.yarn.api.protocolrecords.AllocateResponse;
import org.apache.hadoop.yarn.api.records.FinalApplicationStatus;
import org.apache.hadoop.yarn.client.api.AMRMClient;
import org.apache.hadoop.yarn.client.api.AMRMClient.ContainerRequest;
import org.apache.hadoop.yarn.conf.YarnConfiguration;
import org.apache.hadoop.yarn.exceptions.YarnException;

import javax.net.ssl.SSLException;


//Application Master for stateful-streaming is different from others (Block and windowed).
// Because user code is in full control here, and has access to full HLE-API along with
// Stateful-Streaming API, it will be the user code that will be polling for relevant
// queue sizes, any reservation etc.
// ToDo: How do we stop some user's code from messing with other users' queues if it has access to
// HLE API?  This question ties to overall security model that needs a full review.  There are user
// based access control in underlying HLE API, but at PLUMB level everything runs as HDFS root user.


/**
 * The ApplicationMaster, which is responsible for sending container
 * requests to the ResourceManager, and talking to NodeManagers when
 * the RM provides us with available containers.
 */
public class YARNApplicationMasterForStatefulStreaming {
    private static long t1,t2;

    public static void main(String[] args)  {
        t1 = System.nanoTime();
        int retCode = -1;
        Logger logger = Logger.getAnonymousLogger();
        logger.addHandler(new StreamHandler(System.out, new SimpleFormatter()));
        t2 = System.nanoTime();
        logger.info("time to get loggers = " + (t2-t1)/1000000000);


        t1 = System.nanoTime();
        Map<String, Integer> dupTable = (Map<String, Integer>) MyUtils.deSeralizeObjectFromHDFS(
                ServerConfig.getString("dupPathForAppMaster"),
                ServerConfig.getString("dupPathInHDFS"), logger);


        Map<String, Stage> dispTable = (Map<String, Stage>) MyUtils.deSeralizeObjectFromHDFS(
                ServerConfig.getString("dispPathForAppMaster"),
                ServerConfig.getString("dispPathInHDFS"), logger);

        t2 = System.nanoTime();
        logger.info("time to fetch queuesystem, dup and disp from hdfs = " + (t2-t1)/1000000000);
        t1 = System.nanoTime();



        try {

            Configuration conf = new YarnConfiguration();

            // create a client to talk to the ResourceManager
            AMRMClient<ContainerRequest> rmClient = AMRMClient.createAMRMClient();
            rmClient.init(conf);
            rmClient.start();




            // register with ResourceManager
            System.out.println("registerApplicationMaster: pending");
            rmClient.registerApplicationMaster("", 0, "");
            System.out.println("registerApplicationMaster: complete");


            // sending pings to RM periodically so that it stays alive
            Thread t = new Thread() {
                public void run() {
                    //Random random = new Random();
                    long execTimeInMinutes = 0;

                    while (true) {
                        try {
                            AllocateResponse response = rmClient.allocate(0);
                            TimeUnit.SECONDS.sleep(60);
                            //ToDo:Measuring wall-clock time or duration is not a good idea using
                            //sleep because under load, sleep might fire late.
                            //I tried using System.currentTimeMillis()/(1000*60) but it is performing
                            //worse (e.g. 5 seconds elapses but it shows 0 or 1)
                            execTimeInMinutes += 1;

                            //ToDo: Stateful-streaming should never exit ideally. But I left the following intact
                            // where job will be killed after 4 hours.  This is to force user code to make checkpoint
                            // and recovery operational.  Secondly, if user code is stuck due to some reason, this
                            // App Master will not know about it (Halting problem). To avoid it, and to start with a
                            // clean-slate it is probably better to restart after some time.
                            if(ServerConfig.getInt("maxYARNJobRunTime") < execTimeInMinutes)
                            {
                                System.out.println("Max time to run a YARN job exceeded. Exiting...");
                                System.exit(5);

                            }
                        } catch (YarnException | InterruptedException | IOException e) {
                            e.printStackTrace();
                        }

                        // Following only for testing failure recovery. 10% error rate
                        //if(random.nextInt(100) < 10)
                        //{
                        //    System.exit(2);
                        //}

                    }
                }
            };
            t.start();
            t2 = System.nanoTime();
            logger.info("time to register AM and starting ping thread = " + (t2-t1)/1000000000);


            t1 = System.nanoTime();

            System.out.println("Dump debug output");
            Map<String, String> envs = System.getenv();
            for (Map.Entry<String, String> env : envs.entrySet()) {
                System.out.println("System env: key=" + env.getKey() + ", val=" + env.getValue());

            }

            t2 = System.nanoTime();
            logger.info("time to print environment = " + (t2-t1)/1000000000);



            t1 = System.nanoTime();
            retCode = runIteration(dispTable, dupTable, envs, logger);
            t2 = System.nanoTime();
            logger.info("time to runIteration = " + (t2-t1)/1000000000);







            System.out.println("unregister");
            // Un-register with ResourceManager
            if(retCode == 0)
                rmClient.unregisterApplicationMaster(FinalApplicationStatus.SUCCEEDED, "", "");
            else
                rmClient.unregisterApplicationMaster(FinalApplicationStatus.FAILED, "", "");

            System.out.println("exiting");
        } catch(Throwable t) {
            t.printStackTrace();
        }

        System.exit(retCode);
    }

    private static int run(List<String> args,
                           Stage stage,
                           List<TezQRep> in,
                           List<TezQRep> out) {
        int retCode = 1;
        try {



            args.add(0, "./AppMaster.shellscript");




            ProcessBuilder pb = new ProcessBuilder(args)
                    .inheritIO();


            pb.redirectErrorStream(true);
            setEnv(pb.environment(), stage, in, out);
            Process p = pb.start();
            retCode = p.waitFor();
            BufferedReader reader =  new BufferedReader(new InputStreamReader(p.getInputStream()));
            String line = "";
            while ( (line = reader.readLine()) != null) {
            	System.out.println(line);
            }

            System.out.println("Retval = " + retCode);
        } catch (Exception exp) {
            exp.printStackTrace();
        }

        return retCode;


    }



    private static String getReservedFile(Stage stage, String q, Logger logger) {
        QueueHelper.QResult result;
        //ToDo: The following can become an infinite loop in a scenario, where the file(s) this stage was supposed
        // to do is already done by someone else.
        do {
            result = QueueHelper.reserveTopFileRepeatOnError(q, logger, 1);
            if (result.qSize == 0)
            {
                result.reservedHDFSFile = null;
                break;
            }

        }
        while (result.retCode != 0);

        return result.reservedHDFSFile;
    }




    private static List<String> getCmdLineForYarnv2(Stage stage,
                                                  List<TezQRep> in,
                                                  List<TezQRep> out) {
        List<String> builder = new ArrayList<>();


        for (TezQRep tqr : in) {
            //dont put any spaces in following parameters else getopts
            //in bash fails (probably process builder puts in these spaces
            //in some way that causes getopts to stop
            builder.add(String.format("%s%s", "-i", tqr.hdfsURI));
        }

        for (String s : stage.getProgram()) {
            builder.add(String.format("%s%s", "-p", s));
        }

        for (TezQRep tqr : out) {
            builder.add(String.format("%s%s", "-o", tqr.hdfsURI));
        }

        return builder;
    }




    private static int runIteration(Map<String, Stage> dispTable,
                              Map<String, Integer> dupTable,
                              Map<String, String> envs,
                                     Logger logger)
    {
        t1 = System.nanoTime();


        Stage stage = dispTable.get(envs.get("args_1")); // args_1 should be stage ID

        List<String> inQueues = stage.getInput();
        List<TezQRep> in = new ArrayList<TezQRep>();

        List<String> outQueues = stage.getOutput();
        List<TezQRep> out = new ArrayList<TezQRep>();

        // wait until all input queues have at least one item in it
        for(String q:inQueues)
        {
            TezQRep tqr = new TezQRep(q, q); // for stateful-streaming we don't reserve anything. User code does.
            in.add(tqr);
        }


        for (String q : outQueues) {
            TezQRep tqr = new TezQRep(q, q);
            out.add(tqr);
        }


        List<String> cmdLine = getCmdLineForYarnv2(stage, in, out);
        logger.info("cmdLine = " + Arrays.toString(cmdLine.toArray()));
        t2 = System.nanoTime();
        logger.info("time for file resevations etc. = " + (t2-t1)/1000000000);

        t1 = System.nanoTime();
        int retCode = run(cmdLine, stage, in, out);
        t2 = System.nanoTime();
        logger.info("Return code form run = " + retCode);
        logger.info("time run job = " + (t2-t1)/1000000000);
        t1 = System.nanoTime();

        if(retCode == 0)
        {
            logger.info("Execution successful but stateful-streaming user code exiting...");

        } else
        {
            logger.info("Execution failed with return code " + retCode + "...");
        }
        t2 = System.nanoTime();
        logger.info("time post job run = " + (t2-t1)/1000000000);


        return retCode;

    }


    static class TezQRep {

        public TezQRep(String qName, String hdfsURI) {
            this.qName = qName;
            this.hdfsURI = hdfsURI;
        }

        public String qName;
        public String hdfsURI;
    }



    private static void setEnv(Map<String, String> env,
                               Stage stage,
                               List<TezQRep> in,
                               List<TezQRep> out)
    {

        env.put("__plumb_no_of_inputs", in.size()+"");
        int i = 1;
        for (TezQRep tqr : in)
        {
            env.put("__plumb_i"+i, tqr.hdfsURI);
            i++;
        }



        env.put("__plumb_no_of_outputs", out.size()+"");
        i = 1;
        for (TezQRep tqr : out)
        {
            env.put("__plumb_o"+i, tqr.hdfsURI);
            i++;
        }

        String userProg = "";

        for (String s : stage.getProgram()) {
            userProg += s;
        }

        String parts[] = userProg.split(" ", 2);
        if (null != parts && parts.length > 0)
        {
            env.put("__plumb_prog_name", parts[0]);
        }

        if (null != parts && parts.length > 1)
        {
            env.put("__plumb_prog_args", parts[1]);
        }

        // Put path to the Queue regex.py so that yarn wrapper can sym link to it
        env.put("pathToRegEx", ServerConfig.getString("pathToRegEx"));
        // Also put path for call_lander_api_function_via_rpc.sh
        env.put("pathToHLEBASHAPI", ServerConfig.getString("pathToHLEBASHAPI"));

    }
}
