/*
 * Purpose of this class is to parse and validate YAML ingredients, especially
 * the window.
 */
package edu.isi.ant.yamlseranddeser;

import java.time.Duration;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeParseException;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

/**
 *
 * @author aqadeer
 */
public class Parser {
    
    // https://www.freeformatter.com/java-regex-tester.html#ad-output is a simple site to test regular expressions.
    // Don't use "" or escape characters there.
    
    private static final Pattern sizePattern = Pattern.compile("\\s*(\\d+)\\s*([a-zA-Z]+)\\s*", 
                                                        Pattern.CASE_INSENSITIVE);
    
    private static final String datePat = "\\s*([0-9][0-9][0-9][0-9])\\s*([:\\-_])?([0-9][0-9])\\s*([:\\-_])?([0-9][0-9])\\s*([:\\-_])?([0-9][0-9])\\s*([:\\-_])?([0-9][0-9])\\s*([:\\-_])?([a-zA-Z][a-zA-Z][a-zA-Z])\\s*";
    
    private static final Pattern datePattern = Pattern.compile(datePat);
 
    private static final Pattern startAndEndDatePattern = Pattern.compile(datePat+ "\\s*to\\s*"+datePat, Pattern.CASE_INSENSITIVE);
    
    
    private static final Pattern maxWaitPattern = Pattern.compile("\\s*(\\d+)\\s*([a-zA-Z%]+)\\s*", 
                                                        Pattern.CASE_INSENSITIVE);
    
    private static final Pattern acrossWindowSchedulingPattern = Pattern.compile("\\s*(yes|no)\\s*", 
                                                        Pattern.CASE_INSENSITIVE);
    
    private static final Pattern useQueueSequenceNumberInstancePattern = Pattern.compile("\\s*(yes|no)\\s*", 
                                                        Pattern.CASE_INSENSITIVE);
    
    private static final Pattern queueSequenceNumberInstanceListPattern = Pattern.compile("\\s*(([^,])*,*)*([^,])+\\s*", 
                                                        Pattern.CASE_INSENSITIVE);        
    
    
    //ToDo: To extend windowing so that it also accepts cron string
    //      so that periodic windowing could be scheduled as well like
    //      one hour sampling every day etc.
    
    
        
                                                        
    
    public static Window parseAndEnhance(Map<String,String> windowProp) throws WindowParsingFailedException
    {
        // sanity check
        if(windowProp == null) return null; //if you dump a null on me, I dump it back to you!
        // following 3 for verification that mandatory fields are provided
        boolean size = false;
        boolean start = false;
        boolean startAndEnd = false;
        boolean instanceUsageParameterPresent = false;
        boolean instanceListParameterPresent = false;
        
        Window aWindow = new Window();
        
        for(String key:windowProp.keySet())
        {
            switch(key)
            {
                case "size":
                    SizeAndMetric result = findMetric(windowProp.get(key));
                    if(result.result)
                    {
                        aWindow.sizeMetric = result.metric;
                        aWindow.size = result.size;
                        size=true;
                    }
                    else
                    {
                        onParsingError("Size parsing failed. " + result.errMessage);                        
                    }
                    break;
                    
                case "start":
                    WindowStartTimeResult startResult = findStartOfWindow(windowProp.get(key));
                    if(startResult.result)
                    {
                        aWindow.startOfWindow = startResult.startOfWindow;
                        start=true;
                    }
                    else
                    {
                        onParsingError("Start_of_window parsing failed. " + startResult.errMessage);
                    }
                    break;
                    
                case "start_and_end":
                    WindowStartAndEndTimeResult startAndEndResult = findStartAndEndOfWindow(windowProp.get(key));
                    if(startAndEndResult.result)
                    {
                        aWindow.startOfWindow = startAndEndResult.startOfWindow;
                        aWindow.endOfWindow = startAndEndResult.endOfWindow;
                        aWindow.size = Duration.between(aWindow.startOfWindow,aWindow.endOfWindow).toHours();
                        startAndEnd=true;
                    }
                    else
                    {
                        onParsingError("Start_and_end_of_window parsing failed. " + startAndEndResult.errMessage);                       
                    }
                    break;
                    
                case "max_wait":
                    MaxWaitMetric maxWaitResult = findMaxWait(windowProp.get(key));
                    if(maxWaitResult.result)
                    {
                        aWindow.maxWait = maxWaitResult.waitTimeInHours;
                        aWindow.maxWaitMetric = maxWaitResult.metric;                        
                    }
                    else
                    {
                        onParsingError("Max Wait parsing failed. " + maxWaitResult.errMessage);                        
                    }
                    break;
                    
                case "across_window_ordering":
                    AcrossWindowSchedulingResult schedOptionResult = findAcrossWindowScheduling(windowProp.get(key));
                    if(schedOptionResult.result)
                    {
                        aWindow.outOfOrderScheduling = schedOptionResult.acrossWindowScheduling;                        
                    }
                    else
                    {
                        onParsingError("Across window scheduling request parsing failed. " + schedOptionResult.errMessage);                        
                    }
                    break;
                    
                case "abstraction":
                    aWindow.abstacton = Window.mapStringToAbstraction(windowProp.get(key));
                    break;
                    
                case "cutoff_delivery_requirement":
                    Window.CUTOFFDELIVERYREQUIREMENT req =  Window.mapStringToCutoff(windowProp.get(key));
                    if(req != null)
                    {
                        aWindow.cutoffDeliveryRequirement = req;
                    }
                    else
                    {
                        onParsingError("Window cutoff provided by user not recognized.");                        
                    }                    
                    break;
                    
                case "use_queue_sequence_number_instance":
                    UseQueueSequenceNumberInstanceResult queueInstanceUseResult = findQueueInstanceSeqUse(windowProp.get(key));
                    if(queueInstanceUseResult.result)
                    {
                        aWindow.useQueueSequenceNumberInstance = queueInstanceUseResult.useQueueSequenceNumberInstance;    
                        instanceUsageParameterPresent = true;
                    }
                    else
                    {
                        onParsingError("use_queue_sequence_number_instance parsing failed. " + queueInstanceUseResult.errMessage);                        
                    }
                    break;
                    
                case "queue_sequence_number_instance_list":
                    QueueSequenceNumberInstanceListResult queueInstanceListResult = findQueueSequenceNumberInstanceList(windowProp.get(key));
                    if(queueInstanceListResult.result)
                    {
                        aWindow.queueSequenceNumberInstancesList = queueInstanceListResult.queueSequenceNumberInstancesList;   
                        instanceListParameterPresent = true;
                    }
                    else
                    {
                        onParsingError("queue_sequence_number_instance_list parsing failed. " + queueInstanceListResult.errMessage);                        
                    }
                    break;
                    
                    
                default:
                    onParsingError("Window property " + key + " not supported.");                    
                    break;                        
            }
            
        }
        
        
        // Before returning, need to confirm that all required stuff is in.
        if(!(size) && !(startAndEnd))
        {
            throw new WindowParsingFailedException("Either [size] or [start and end] are required for a window.");
        }
        // But not both of them
        if((size && start) && (startAndEnd))
        {
            throw new WindowParsingFailedException("Either provide [size and start] or [start and end].  Not both!");
        }
        // If user didn't provide: use_queue_sequence_number_instance but did provide: queue_sequence_number_instance_list
        // then toggle first one to Yes because default is No
        if (instanceListParameterPresent && !instanceUsageParameterPresent)
        {
            aWindow.useQueueSequenceNumberInstance = Window.USEQUEUESEQUENCENUMBERINSTANCE.Yes;
        }
        aWindow.timeMetric = ((size && start)?Window.TIMEMETRIC.TimeViaSizeAndStart:Window.TIMEMETRIC.TimeViaStartAndEnd);
        return aWindow;       
    }
    
    
    private static void onParsingError(String msg) throws WindowParsingFailedException
    {
        throw new WindowParsingFailedException(msg);
    }
    
    public static SizeAndMetric findMetric (String size)
    {
        // 24 hours  or 24 Hours   or 24hours
        // 100 blocks or 100 block or 100 Blocks or 100Blocks
        // Two groups (digit)(text)
        Matcher matcher = sizePattern.matcher(size);
        SizeAndMetric result = new SizeAndMetric();
        if(!matcher.matches())
        {
            result.result = false;
            result.errMessage = String.format("Given string (%s) does not match required format.", size);
        }
        else
        {
            result.size = Integer.parseInt(matcher.group(1));
            result.result = true;
            result.errMessage = "OK";
            if(matcher.group(2).toLowerCase().contains("block"))
            {
                result.metric = Window.SIZEMETRIC.Blocks;
            }
            else if(matcher.group(2).toLowerCase().contains("hour"))
            {
                result.metric = Window.SIZEMETRIC.Time;
            }
            else
            {
                result.result = false;
                result.errMessage = "window size can be in terms of hours, or blocks only.  Found none of them.";                
            }
            

        }
        
        return result;        
    }
    
    
    public static WindowStartTimeResult findStartOfWindow (String startTime)
    {
            // YYYYMMDDhhmmUTC or YYYY:MM:DD:hh:mm:UTC or or YYYY:MM:DD:hh:mm UTC
            // or YYYY-MM-DD-hh-mm-UTC  or YYYY_MM_DD_hh_mm_UTC
        Matcher matcher = datePattern.matcher(startTime);
        WindowStartTimeResult result = new WindowStartTimeResult();
        if(!matcher.matches())
        {
            result.result = false;
            result.errMessage = String.format("Given time (%s) does not match required format (YYYY:MM:DD:hh:mm UTC).", startTime);
        }
        else
        {
            String reArrangeStartTime = String.format("%s %s %s %s %s %s", matcher.group(1), matcher.group(3), matcher.group(5),
                                                        matcher.group(7),matcher.group(9),matcher.group(11));
            
            try{
                // convert user provided time zone to UTC
                //result.startOfWindow = ZonedDateTime.parse(reArrangeStartTime, DateTimeFormatter.ofPattern("yyyy MM dd HH mm zzz")).withZoneSameInstant(ZoneOffset.UTC);
                // See https://stackoverflow.com/questions/39506891/why-is-zoneoffset-utc-zoneid-ofutc why below gives us our required result instead of above one
                result.startOfWindow = ZonedDateTime.parse(reArrangeStartTime, DateTimeFormatter.ofPattern("yyyy MM dd HH mm zzz")).withZoneSameInstant(ZoneId.of("UTC"));
                result.result = true;
                result.errMessage = "OK";
            }catch (DateTimeParseException exp)
            {
                //exp.printStackTrace();
                result.result = false;
                result.errMessage = String.format("[DateTimeParseException]Given time (%s) does not match required format (YYYY:MM:DD:hh:mm UTC).", reArrangeStartTime);
                
            }
        }
        
        return result;        
    }
    
    
    public static WindowStartAndEndTimeResult findStartAndEndOfWindow (String startAndEndTime)
    {
        // YYYYMMDDhhmmUTC to YYYYMMDDhhmmUTC
        Matcher matcher = startAndEndDatePattern.matcher(startAndEndTime);
        WindowStartAndEndTimeResult result = new WindowStartAndEndTimeResult();
        if(!matcher.matches())
        {
            result.result = false;
            result.errMessage = String.format("Given start and end time (%s) does not match required format (YYYY:MM:DD:hh:mm UTC to YYYY:MM:DD:hh:mm UTC).", startAndEndTime);
        }
        else
        {
            String reArrangeStartTime = String.format("%s %s %s %s %s %s", matcher.group(1), matcher.group(3), matcher.group(5),
                                                        matcher.group(7),matcher.group(9),matcher.group(11));
            
            try{
            result.startOfWindow = ZonedDateTime.parse(reArrangeStartTime, DateTimeFormatter.ofPattern("yyyy MM dd HH mm zzz")).withZoneSameInstant(ZoneId.of("UTC"));;           
            result.result = true;
            result.errMessage = "OK";
            }catch (DateTimeParseException exp)
            {
                //exp.printStackTrace();
                result.result = false;
                result.errMessage = String.format("[DateTimeParseException]Given time (%s) does not match required format (YYYY:MM:DD:hh:mm UTC).", reArrangeStartTime);
                
            }
            
            String reArrangeEndTime = String.format("%s %s %s %s %s %s", matcher.group(12), matcher.group(14), matcher.group(16),
                                                        matcher.group(18),matcher.group(20),matcher.group(22));
            
            try{
                //convert following user provided time zone to UTC
                result.endOfWindow = ZonedDateTime.parse(reArrangeEndTime, DateTimeFormatter.ofPattern("yyyy MM dd HH mm zzz")).withZoneSameInstant(ZoneId.of("UTC"));
                //result.endOfWindow = ZonedDateTime.parse(reArrangeEndTime, DateTimeFormatter.ofPattern("yyyy MM dd HH mm zzz")).withZoneSameInstant(ZoneOffset.UTC);  
                //result.endOfWindow = ZonedDateTime.parse(reArrangeEndTime, DateTimeFormatter.ofPattern("yyyy MM dd HH mm zzz"));
                result.result = true;
                result.errMessage = "OK";
            }catch (DateTimeParseException exp)
            {
                //exp.printStackTrace();
                result.result = false;
                result.errMessage += String.format("[DateTimeParseException]Given time (%s) does not match required format (YYYY:MM:DD:hh:mm UTC).", reArrangeStartTime);
                
            }
            
            if(result.endOfWindow.isBefore(result.startOfWindow) || result.endOfWindow.isEqual(result.startOfWindow))
            {
                result.result = false;
                result.errMessage = "Window end date is eiter equal or before than start date";
            }
        }
        
        return result;        
    }
    
    
    
    public static MaxWaitMetric findMaxWait (String maxWait)
    {
        // xhours like 1hour or 1 Hour or 0.5 Hour 
        // yblocks or y Block
        // 2% or 2 percent
        Matcher matcher = maxWaitPattern.matcher(maxWait);
        MaxWaitMetric result = new MaxWaitMetric();
        if(!matcher.matches())
        {
            result.result = false;
            result.errMessage = String.format("Given string (%s) does not match required format.", maxWait);
        }
        else
        {
            result.result = true;
            result.errMessage = "OK";
            
            
            result.waitTimeInHours = Integer.parseInt(matcher.group(1));
            if(matcher.group(2).toLowerCase().contains("block"))
            {
                result.metric = Window.MAXWAITEMETRIC.Blocks;
            }
            else if(matcher.group(2).toLowerCase().contains("hour"))
            {
                result.metric = Window.MAXWAITEMETRIC.Time;
            }
            else if(matcher.group(2).toLowerCase().contains("percent") || matcher.group(2).contains("%") )
            {
                result.metric = Window.MAXWAITEMETRIC.PercentBlocks;
            }
            else
            {
                result.result = false;
                result.errMessage = "max_wait can either in terms of hours, or number or blocks of % blocks.  Found none of them.";                
            }
            

        }
        
        return result;        
    }
    
    
    public static AcrossWindowSchedulingResult findAcrossWindowScheduling (String schedOption)
    {
        // yes or no

        Matcher matcher = acrossWindowSchedulingPattern.matcher(schedOption);
        AcrossWindowSchedulingResult result = new AcrossWindowSchedulingResult();
        if(!matcher.matches())
        {
            result.result = false;
            result.errMessage = String.format("Given string (%s) does not match required format.", schedOption);
        }
        else
        {
            
            result.result = true;
            result.errMessage = "OK";

            if(matcher.group(1).toLowerCase().contains("yes"))
            {
                result.acrossWindowScheduling = Window.OUTOFORDERSCHEDULING.Yes;
            }
            else
            {
                result.acrossWindowScheduling = Window.OUTOFORDERSCHEDULING.No;                
            }
            

        }
        
        return result;        
    }
    
    public static UseQueueSequenceNumberInstanceResult findQueueInstanceSeqUse (String userString)
    {
        // yes or no

        Matcher matcher = useQueueSequenceNumberInstancePattern.matcher(userString);
        UseQueueSequenceNumberInstanceResult result = new UseQueueSequenceNumberInstanceResult();
        if(!matcher.matches())
        {
            result.result = false;
            result.errMessage = String.format("Given string (%s) does not match required format.", userString);
        }
        else
        {
            
            result.result = true;
            result.errMessage = "OK";

            if(matcher.group(1).toLowerCase().contains("yes"))
            {
                result.useQueueSequenceNumberInstance = Window.USEQUEUESEQUENCENUMBERINSTANCE.Yes;
            }
            else
            {
                result.useQueueSequenceNumberInstance = Window.USEQUEUESEQUENCENUMBERINSTANCE.No;                
            }
            

        }
        
        return result;        
    }
    
    
    public static QueueSequenceNumberInstanceListResult findQueueSequenceNumberInstanceList (String userString)
    {
        // a comma separated list

        Matcher matcher = queueSequenceNumberInstanceListPattern.matcher(userString);
        QueueSequenceNumberInstanceListResult result = new QueueSequenceNumberInstanceListResult();
        if(!matcher.matches())
        {
            result.result = false;
            result.errMessage = String.format("Given string (%s) does not match required format.", userString);
        }
        else
        {
            
            result.result = true;
            result.errMessage = "OK";
            result.queueSequenceNumberInstancesList = userString;           

        }
        
        return result;        
    }
    
        
    
    
    
    // Some testing
    public static void main(String[] args)
    {
        //testWindowCutoffParsing();
        //testUseQueueSequenceNumberInstanceParsing();
        testQueueSequenceNumberInstanceListParsing();
              
    }
    
    private static void testSizeParsing()
    {
        String sizeTestArray[] = {"24 hours", "24hours", "100 blocks", "100block", "bad", "30 minutes"};
        for (String testString : sizeTestArray) {
            SizeAndMetric result = Parser.findMetric(testString);
            System.out.println(result);

        }
        
    }
    
    private static void testMaxWaitParsing()
    {
        String sizeTestArray[] = {"1 hours", "1hours", "100 blocks", "100block", "1 month", "1%", "1 percent"};
        for (String testString : sizeTestArray) {
            MaxWaitMetric result = Parser.findMaxWait(testString);
            System.out.println(result);

        }
        
    }
    
    private static void testStartOfWindowParsing()
    {
        String testArray[] = {"202004010000UTC", "2020:04:01:00:00UTC", "2020 04 01 00 00 UTC", "2020-04-01-00-00PDT", "bad"};
        
        for (String testString : testArray) {
            WindowStartTimeResult result = Parser.findStartOfWindow(testString);
            System.out.println(result);

        }
        
    }
    
    
    private static void testStartAndEndOfWindowParsing()
    {
        String testArray[] = {"202004010000UTC to 202004020000UTC", 
                              "202004010000UTC To 202004020000UTC", "202004010000UTC", 
                              "202004010000UTC to 202004010000UTC", "202004010000UTC to 202004010000PDT"};
        
        for (String testString : testArray) {
           WindowStartAndEndTimeResult result = Parser.findStartAndEndOfWindow(testString);             
            System.out.println(result);

        }
        
    }
    
    
        private static void testAbstractonParsing()
    {
        String testArray[] = {"LB_WINDOW_UNORDERED_MapReduce", "xyz"};
        
        for (String testString : testArray) {                        
            System.out.println(Window.mapStringToAbstraction(testString));
        }
        
    }
    
    
    
    private static void testAcrossWindowSchedulingParsing()
    {
        String testArray[] = {"yes", "No", "bad"};
        
        for (String testString : testArray) {
            AcrossWindowSchedulingResult result = Parser.findAcrossWindowScheduling(testString);             
            System.out.println(result);

        }
        
    }
    
    //
    private static void testWindowCutoffParsing()
    {
        String testArray[] = {"Cutoff_Exactly_Once_To_Current_Window", "Dont cut"};
        
        for (String testString : testArray) {
            Window.CUTOFFDELIVERYREQUIREMENT req =  Window.mapStringToCutoff(testString);             
            System.out.println(req);

        }
        
    }
    
    private static void testUseQueueSequenceNumberInstanceParsing()
    {
        String testArray[] = {"yes", "No", "bad"};
        
        for (String testString : testArray) {
            UseQueueSequenceNumberInstanceResult result = Parser.findQueueInstanceSeqUse(testString);             
            System.out.println(result);

        }
        
    }
    
    
    private static void testQueueSequenceNumberInstanceListParsing()
    {
        String testArray[] = {"lax", "lax,mia", "lax, mia", "lax,ari,mia,"};
        
        for (String testString : testArray) {
            QueueSequenceNumberInstanceListResult result = Parser.findQueueSequenceNumberInstanceList(testString);             
            System.out.println(result);

        }
        
    }   
}

    class SizeAndMetric
    {
        public int size;
        public Window.SIZEMETRIC metric;
        
        public boolean result;
        public String errMessage;
        
        public  String toString()
        {
            return String.format("size = %d  size metric = %s   result = %b   errMessage = %s", 
                                  size, metric, result, errMessage);            
        }
    }

    class MaxWaitMetric
    {
        public int waitTimeInHours;
        public Window.MAXWAITEMETRIC metric;
        
        public boolean result;
        public String errMessage;
        
        public  String toString()
        {
            return String.format("waitTimeInHours = %d  wait metric = %s   result = %b   errMessage = %s", 
                                  waitTimeInHours, metric, result, errMessage);            
        }
    }

    class WindowStartTimeResult
    {
        public ZonedDateTime startOfWindow;
         
        public boolean result;
        public String errMessage;
        
        public  String toString()
        {
            return String.format("Window start time:  = %s    result = %b   errMessage = %s", 
                                  startOfWindow, result, errMessage);            
        }
    }

    class WindowStartAndEndTimeResult
    {
        public ZonedDateTime startOfWindow;
        public ZonedDateTime endOfWindow;
         
        public boolean result;
        public String errMessage;
        
        public  String toString()
        {
            return String.format("Window start time:  = %s    Window end time:  = %s result = %b   errMessage = %s", 
                                  startOfWindow, endOfWindow, result, errMessage);            
        }
    }


    class AcrossWindowSchedulingResult
    {
        public Window.OUTOFORDERSCHEDULING acrossWindowScheduling;
         
        public boolean result;
        public String errMessage;
        
        public  String toString()
        {
            return String.format("across window scheduling:  = %s   result = %b   errMessage = %s", 
                                  acrossWindowScheduling, result, errMessage);            
        }
    }

    class UseQueueSequenceNumberInstanceResult
    {
        public Window.USEQUEUESEQUENCENUMBERINSTANCE useQueueSequenceNumberInstance;
         
        public boolean result;
        public String errMessage;
        
        public  String toString()
        {
            return String.format("use_queue_sequence_number_instance:  = %s   result = %b   errMessage = %s", 
                                  useQueueSequenceNumberInstance, result, errMessage);            
        }
    }

    class QueueSequenceNumberInstanceListResult
    {
        public String queueSequenceNumberInstancesList;
         
        public boolean result;
        public String errMessage;
        
        public  String toString()
        {
            return String.format("queue_sequence_number_instance_list:  = %s   result = %b   errMessage = %s", 
                                  queueSequenceNumberInstancesList, result, errMessage);            
        }
    }
/*
What was tested:
(1) //ToDo: If parsing of ANY field fails, throw and exception and stop

(2) //ToDo: Verify mandatory fields of the window are in
    [Done]



*/
