#!/bin/bash

set -e

KEY_BITS=4096
KEY_DAYS=365

usage() {
cat <<EOF
$0: set up and manage certificates for HLE communication through proxy

USAGE:
	$0 dir [setup|create <key_basename> <common_name>]

EXAMPLES:
- create CA certificates, keys; used by LANDER admin for setup:
	$0 wip/test/keys/hcpu001.ant.isi.edu setup

- create certificates for running proxy on hcpu001.ant.isi.edu:
	$0 wip/test/keys/hcpu001.ant.isi.edu create server hcpu001.ant.isi.edu

- create certificates for user 'yuri':
	$0 wip/test/keys/hcpu001.ant.isi.edu create yuri yuri
EOF
exit $1
}

setupCA() {
	local _d=$1 _f
	(
		cd $_d;
		_f="$PWD/ca.key"
		[ -e "$_f" ] && {
			echo "WARNING: file $_f exists, please remove $PWD/ca.*, $PWD/roots.pem before running this command" >&2
			exit 1
		}
		_f="$PWD/ca.crt"
		[ -e "$_f" ] && {
			echo "WARNING: file $_f exists, please remove $PWD/ca.*, $PWD/roots.pem before running this command" >&2
			exit 1
		}
		_f="$PWD/roots.pem"
		[ -e "$_f" ] && {
			echo "WARNING: file $_f exists, please remove $PWD/ca.*, $PWD/roots.pem before running this command" >&2
			exit 1
		}
		umask 0077
		openssl genrsa -des3 -out ca.key $KEY_BITS
		openssl req -new -x509 -days $KEY_DAYS -key ca.key -out ca.crt -subj "/CN=lander:$(hostname)"
		cp ca.crt roots.pem
	) || exit 1
	exit 0
}

createCRT() {
	local _d=$1 _b=$2 _cn=$3
	(
		cd $_d
		openssl genrsa -out $_b.key $KEY_BITS
		openssl req -new -key $_b.key -out $_b.csr -subj "/CN=$_cn"
		openssl x509 -req -days $KEY_DAYS -in $_b.csr -CA ca.crt -CAkey ca.key -set_serial 01 -out $_b.crt
		openssl pkcs8 -topk8 -inform PEM -outform PEM -nocrypt -in $_b.key -out $_b.key.pem
		cat $_b.crt ca.crt > ${_b}_chain
	) || exit 1
    exit 0
}


###MAIN

[ "$#" -lt 2 ] && usage 1 >&2
 
dir=$1
cmd=$2

shift 2

[ -d "$dir" ] || {
	echo "ERROR: directory $dir does not exist" >&2
	exit 1;
}

case "$cmd" in
	( "setup"  ) setupCA   $dir $@ ;;
	( "create" ) createCRT $dir $@ ;;
	( * ) usage 1 >&2;
esac

#not reached
exit 1 
