#!/usr/bin/env python3

import random
import sys

from top.sup import opsys, db
from top.conf import sysConfInfo, syssetup




Host =  sysConfInfo.Host
User =  sysConfInfo.User
Pass =  sysConfInfo.Pass
DB   =  sysConfInfo.DB


debug = True

# Assumptions:
# (1) userName (i.e. queue_owner) should be added to Linux group data_type
# (2) data_type and corresponding Linux groups is currently verified from db.
#     It is expected that via sysInit.py, these values have been loaded from sysConfInfo file


# Weaknesses
# (1) Each time this function is called, a new db connectoin and cursor is acquired.  Such info
#     can be cached somewhere and then re-used.  But doing so complicates code in terms of sharing
#     between mutliple concurrent calls.  We commit at different times for different calls, and managing
#     that with shared connections is way too complicated.
#
# (2) If queue creation call fails, it can leave zombie tables in the database and possibly zombie directories
#     inside HDFS.  A solution to handel this problem is to implement concept of transactions on both HDFS and
#     db.  Dosing so is not trivial and is not done yet.  Another possible solution could be to see exact return
#     code in createQueue funciton and if failures are of certain type, then call deleteQueue().  But doing so is
#     not robust (e.g. if error codes change due to code change, code breaks!)

def createQueue_Inner(queue_id, data_type, userName):
    
    dbcon = db.__connectToDB(Host, User, Pass, DB)
    if dbcon == None:
        return 1, "dbcon failed inside %s" %(__name__)
    try:
        cur = dbcon.cursor()
    except MySQLdb.Error as e:
        return 2, "cursor acquistion failed inside %s" %(__name__)
        
    dbcon.autocommit(False)

    # Confirm system knowns about data_type

    validy, groupName = db.__getValFromKeyValTable(dbcon, cur, data_type)
    if validy != 0:
        return 3, "failed to verify validity of data_type inside %s" %(__name__)


    if debug: print ("userName = "  + userName)
    if debug: print ("groupName = " + groupName)

    # If calling user is admin, by-pass following check. (Needed it for plumb-tool mk-q)
    if userName.lower() != "admin" and opsys.__doesUserBelongToGroup(userName, groupName) != 0:
        return 4, "user %s does not belong to group %s.  This failure is in %s" %(userName, groupName, __name__)


    if syssetup.api_prepareHDFSDir(data_type, groupName) != 0:
        return 5, "failed to create appropriate HDFS directory structure.  This failure is in %s" %(__name__)


    db_user = userName

    if db.__doesQueueExists(dbcon, cur, queue_id) != 1:
        return 0, "queue with id %s already exists.  Rememebr queue_id are unique across the system.  This failure is in %s" %(queue_id, __name__)

    # XXX following two db interactions should be a transaction BUT
    # DDL operations like table creation are not top.supported in
    # transaction in MySQL


    if db.__addQueue(dbcon, cur, queue_id, db_user, False) != 0:
        return 7, "db.__addQueue failed that was called from %s" %(__name__)
    

    if db.__populateQueueTable(dbcon, cur, queue_id, data_type, db_user, False) != 0:
        sys.stderr.write("createQueue:::Queue table insertion failed.\n")
        return 8, "db.__populateQueueTable failed that was called from  %s" %(__name__)

    

    # if everything went well only now is the time to commit
    dbcon.commit()
    cur.close()
    db.__disconnectFromDB(dbcon)

    
    return 0, "OK"

#..........................................................................
def createQueue(queue_id, data_type, queue_owner):
    retVal, retMessage = createQueue_Inner(queue_id, data_type, queue_owner)
    print (retVal)
    print (retMessage)
    return retVal, retMessage

#..........................................................................
if __name__ == '__main__':
    queue_id = "queue_2"
    data_type = "raw"
    queue_owner = "someboday"
    createQueue(queue_id, data_type, queue_owner)


