#!/usr/bin/env python3

import os
import sys
import stat

from .top.sup import opsys, db, hdfs
from .top.conf import sysConfInfo
from .top.queue_admin import listDerivedQueues
from .top.sup import regex

Host =  sysConfInfo.Host
User =  sysConfInfo.User
Pass =  sysConfInfo.Pass
DB   =  sysConfInfo.DB

debug = True

def e(ret):
    exit(ret[0])
	
def __doDirectDeliveryToQueue(dbcon, cur, hdfsURI, modTime, src_queue_id, queue_owner, eventTime, eventTimeFormat, \
                              sequence_no, optional_info, sequence_no_instance, tailQID, shouldCommit):
    if db.__populateDataTable(dbcon, cur, hdfsURI, src_queue_id, False) != 0:
        return (20, "db.__populateDataTable failed.  This failure is in %s" 
                      %(__name__))
    
    fileState = "UNRESERVED"    
    if db.__populateSubscriberQueueTable(dbcon, cur, src_queue_id, queue_owner, 
                                         hdfsURI, fileState, modTime, eventTime, eventTimeFormat, sequence_no, optional_info, sequence_no_instance, tailQID, False) != 0:
        return (21, "db.__populateSubscriberQueueTable failed.  This failure is in %s" 
                      %(__name__))
    
    # Now lets try to commit the transaction
    try:
        if shouldCommit == True:
            dbcon.commit()	    
        return (0, "OK")
    except MySQLdb.Error as e:
        if debug: print (e)
        return (22, "DB commit failure.  This failure is in %s" 
                      %(__name__))    
    

def enqueueLocal_Phase1(to_queue_id, localfile, queue_owner):
    dbcon = db.__connectToDB(Host, User, Pass, DB)
    if dbcon == None:
        return (1, "dbcon failed inside %s" %(__name__), None)
    try:
        cur = dbcon.cursor()
    except MySQLdb.Error as e:
        return (2, "cursor acquisition failed inside %s" %(__name__), None)

    if db.__doesQueueExists(dbcon, cur, to_queue_id) != 0:
        return (3, "queue with id %s does not exist.  This failure is in %s" %
                     (to_queue_id, __name__), None)

    # Confirm queue_owner is truely owner of to_queue_id
    (retCode, queue_owner_fromDB) = db.__getQueueOwner(dbcon, cur, to_queue_id)
    if retCode != 0 or queue_owner != queue_owner_fromDB:
        return (4, "%s is not owner of queue %s. This failure is in %s" %
                     (queue_owner, to_queue_id, __name__), None)

    # Direct delivery is only possible if to_queue is active
    (isSuccess, val) = db.__getValFromSubscriberKeyValTable(dbcon, cur, to_queue_id, 
                                                            queue_owner, "isActive")
    if isSuccess != 0 or val != "1":
        return (5, "Queue %s is not active and hence can't receive data.  This failure is in %s"
                     %(to_queue_id, __name__), None)

    # find what is HDFS base dir
    (isSuccess, hdfsBaseDir) = db.__getValFromKeyValTable(dbcon, cur, "HDFS_Base_Dir")
    hdfsURI = ""
    data_type = ""
    if isSuccess != 0:
        return (6, "Failed to get HDFS_Base_Dir from Db.  This failure is in %s " 
                   %(__name__), None)

    # find data_type associated with the to_queue_id
    (isSuccess, data_type) = db.__getDataTypeFromQueueTable(dbcon, cur, to_queue_id)
    if isSuccess != 0:
        return (7, "Failed to get data_type associated with queue %s.  This failure is in %s " 
                     %(to_queue_id, __name__), None)

    # find Linux group associated with data_type
    (isSuccess, groupName) = db.__getValFromKeyValTable(dbcon, cur, data_type)
    if isSuccess != 0:
        return (8, "Failed to get Linux group associated with data_type %s.  This failure is in %s"
                     %(data_type, __name__), None)

    # lets make hdfs_uri
    dstHdfsURI =  "%s/%s/%s/%s" %(hdfsBaseDir, data_type, to_queue_id, os.path.basename(localfile))
    dstHdfsDir =  "%s/%s/%s/" %(hdfsBaseDir, data_type, to_queue_id)

    # make dir hierarchy if does not exist
    if hdfs.makeHDFSDirIfDoesNotExist(dstHdfsDir) != 0:
        return (9, "Failed to make HDFS directory hierarchy %s.  This failure is in %s"
                     %(dstHdfsDir, __name__), None)

    # protect directory
    result = hdfs.setFileOrDirRights(dstHdfsDir, sysConfInfo.DirRights)
    if result != 0:
        return (10, " Failed to set dir rights.  This failure is in %s " %(__name__), None)

    result = hdfs.setFileOrDirGroup(dstHdfsDir, groupName)
    if result != 0:
        return (11, " Failed to set dir group.  This failure is in %s " %(__name__), None)

    cur.close()
    db.__disconnectFromDB(dbcon)

    return (0, "OK", dstHdfsURI)


def enqueueLocal_Phase2(to_queue_id, localfile, hdfs_file_uri, queue_owner,
                        eventTime, eventTimeFormat, sequence_no, optional_info, sequence_no_instance, tailQID):
    # some checks from phase1 needs repeatition because we don't want to be in a situation
    # where client only does round two of call and something changes between two rounds
    retCode, errMessage, dstHdfsURI = enqueueLocal_Phase1(to_queue_id, localfile, queue_owner)

    if retCode != 0: return (retCode, errMessage, dstHdfsURI)

    if dstHdfsURI != hdfs_file_uri:
        return (12, "Client provided uri %s is different from what system asked for %s" %
                      (hdfs_file_uri, dstHdfsURI), None)

    # protect file
    result = hdfs.setFileOrDirRights(hdfs_file_uri, sysConfInfo.FileRights)
    if result != 0:
        return (13, "Failed to set file rights on file %s.  This failure is in %s " %
                      (hdfs_file_uri, __name__), None)


    result, modTime = hdfs.getHDFSFileModTime(hdfs_file_uri)
    if result != 0:
        return (14, "Failed to get mod time of HDFS file %s.  This failure is in %s" %
                      (hdfs_file_uri, __name__), None)


    dbcon = db.__connectToDB(Host, User, Pass, DB)
    if dbcon == None:
        return 1, "dbcon failed inside %s" %(__name__), None
    try:
        cur = dbcon.cursor()
    except MySQLdb.Error as e:
        return 2, "cursor acquisition failed inside %s" %(__name__), None


    result, errMessage = __doDirectDeliveryToQueue(dbcon, cur, hdfs_file_uri, modTime, 
                                                   to_queue_id, queue_owner, eventTime, eventTimeFormat, sequence_no, optional_info, sequence_no_instance, tailQID, True)
    
    cur.close()
    db.__disconnectFromDB(dbcon)

    return (result, errMessage, None)



def enqueueLocal(request, phaseNo):
    (retVal, retMessage, dstHdfsURI) = (30, "Empty", "")
    # enforece our naming conventoins
    retCode, retMessage, date, sequence_no, optional_info, sequence_no_instance, tail = \
        regex.match_file_name(os.path.basename(request.localfile), request.to_queue_id)
    if 0 != retCode:
        return (32, retMessage, "")



    if phaseNo == 1:
      (retVal, retMessage, dstHdfsURI) = enqueueLocal_Phase1(request.to_queue_id, request.localfile, request.caller_id)
    elif phaseNo == 2:
      (retVal, retMessage, dstHdfsURI) = enqueueLocal_Phase2(request.to_queue_id, request.localfile, request.hdfs_file_uri, \
                                         request.caller_id, date, regex.get_event_time_format(), sequence_no, optional_info, sequence_no_instance, tail)
    else:
      (retVal, retMessage, dstHdfsURI) = (31, "enqueueLocal has exactly 2 rounds.", "")

    print (retVal)
    print (retMessage)
    if dstHdfsURI is not None:
      print (dstHdfsURI)
    else:
      dstHdfsURI = ""
    
    return (retVal, retMessage, dstHdfsURI)


if __name__ == '__main__':
    localfile = "/nfs/ruby/aqadeer/working/traces/20141117-065927-00358611.erf.sz"
    to_queue_id = "queue_2"
    queue_owner = "abc"
    phaseNo = 1
    
    #retVal =  enqueueLocal(to_queue_id, localfile, queue_owner, phaseNo)
    #print ("retVal = " + str(retVal))

    # ToDo:  Add local test code as per 2 phase changes

