import inspect
import os
import time
from concurrent import futures

import grpc

from top.conf import sysConfInfo
from top.sup import queue_api_pb2
from top.sup import queue_api_pb2_grpc
from queue_api_pb2 import WindowOperationResponse

KeyPath = os.path.expanduser(sysConfInfo.KeyPath)

# HEADS UP 1: gRPC doesn't like to put None values into primitive types
# and it consider String a primitive type.  So make sure you don't put
# any None for a String in the response object.  Else there will be a
# hard to debug exception on the client side!
#
# HEADS UP 2: Thoroughly test all server side code as in unit test because
# for most errors, the client side will show weired / hard to pin-point
# error.


from top.queue_admin import listDerivedQueues as ldq
from top.queue_admin import createQueue as cq
from top.queue_admin import deleteQueue as dq
from top.queue_admin import connectDerivedQueue as cdq
from top.queue_admin import activateQueue as aq
from top.queue_admin import deactivateQueue as daq
from top.queue_admin import markAsFaulty as maf
from top.queue_admin import markAsHealthy as mah
from top.queue_admin import lander_queue_du as lqd
from top.queue_admin import lander_queue_user_du as lqud
from top.queue_admin import lander_queue_queue2user as lqq2u
from top.queue_admin import lander_queue_user2queues as lqu2q
from top.queue_admin import doErrorRecovery as der
from top.queue_admin import deactivateAllQueues as dallq

from top.queue_push import enqueueLocal as enq
from top.queue_push import distributeLocal as distloc

from top.queue_pull import lander_queue_size as lan_q_size
from top.queue_pull import lander_queue_files as lan_q_files
from top.queue_pull import lander_queue_file_older as lan_q_file_old
from top.queue_pull import lander_queue_reserve_top as lan_q_res_top
from top.queue_pull import lander_queue_reserve_file as lan_q_res_file
from top.queue_pull import lander_queue_unreserve as lan_q_unres
from top.queue_pull import lander_queue_release as lan_q_rel
from top.queue_pull import lander_queue_release_unconditional as lan_q_rel_uncon
from top.queue_pull import lander_queue_wipe as lan_q_wip
from top.queue_pull import listFilesInStateX as lfisx
from top.queue_pull import lander_queue_list as lql
from top.queue_pull import lander_queue_peek_next as peek_next
from top.queue_pull import lander_queue_reserve_next as res_next
from top.queue_pull import lander_queue_list_keys as lan_q_ls_keys
from top.queue_pull import get_value as gv
from top.queue_pull import put_value as pv
from top.queue_pull import del_value as dv

from top.sup.regex import match_file_name as mfn

_ONE_DAY_IN_SECONDS = 60 * 60 * 24

from top.queue_admin import windowmanager

wm = None

admin_only_functions = {"deactivateAllQueues": 0, "startMaintainingWindowState": 1, "getWindowsToSchedule": 2,
                        "changeWindowState": 3, "ls_windowed_queues": 4, "ls_all_window_statuses": 5, "ls_win": 6,
                        "ls_win_in_status": 7, "set_win_status": 8, "ls_f_in_win": 9, "mv_f_in_win": 10,
                        "ls_missing_files_in_a_window": 11, "ls_site_info_in_a_window": 12}


class LanderQueueAPIImplementation(queue_api_pb2_grpc.LanderQueueAPIImplementationServicer):

    def RuleChecking(self, request, context):
        apiFunction = inspect.currentframe().f_back.f_code.co_name
        # thiFunction = inspect.currentframe().f_code.co_name
        print("Got call from %s" % apiFunction)

        caller_id_from_certificate_as_bytes = context.peer_identities()[0]
        caller_id_from_certificate = caller_id_from_certificate_as_bytes.decode('utf-8')
        print("Caller = %s and  claimed caller = %s" % (caller_id_from_certificate, request.caller_id))

        global admin_only_functions
        if apiFunction in admin_only_functions and caller_id_from_certificate.title() != 'Admin':
            return 3, "In API call %s, caller is not Admin" % apiFunction

        if True:
            if caller_id_from_certificate.title() == 'Admin' and request.caller_id == "":
                return 1, "In API call %s, caller is Admin but caller_id argument is not set" % apiFunction
            if caller_id_from_certificate.title() != 'Admin' and request.caller_id != "":
                return 2, "In API call %s, caller is not Admin and caller_id argument is set" % apiFunction

            if caller_id_from_certificate.title() != 'Admin':
                request.caller_id = caller_id_from_certificate

        return 0, "OK"

    def SayHello(self, request, context):
        print(context.peer_identities())
        return queue_api_pb2.HelloReply(message='Hello, %s!' % request.name)

    def SayHelloAgain(self, request, context):
        return queue_api_pb2.HelloReply(message='Hello again, %s!' % request.name)

    def SayTwoWayStreamingHello(self, request_iterator, context):
        print(context.peer_identities())
        i = 1
        for request in request_iterator:
            yield queue_api_pb2.HelloReply(message="Hello %s, %s!" % (str(i), request.name))
            i = i + 1

    def listDerivedQueues(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage, derivedQueues = ldq.listDerivedQueues(request.from_queue_id,
                                                                                 request.caller_id)
        if reply.retCode == 0:
            reply.result = True
            reply.listOfDerivedQueues.extend(derivedQueues)

        return reply

    def createQueue(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage = cq.createQueue(request.queue_id, request.data_type, request.caller_id)
        if reply.retCode == 0: reply.result = True
        return reply

    def deleteQueue(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage = dq.deleteQueue(request.queue_id, request.caller_id)
        if reply.retCode == 0: reply.result = True
        return reply

    def connectDerivedQueue(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage = cdq.connectDerivedQueue(request.from_queue_id, request.to_queue_id,
                                                                    request.caller_id)
        if reply.retCode == 0: reply.result = True
        return reply

    def activateQueue(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage = aq.activateQueue(request.queue_id, request.caller_id)
        if reply.retCode == 0: reply.result = True
        return reply

    def deactivateQueue(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage = daq.deactivateQueue(request.queue_id, request.caller_id)
        if reply.retCode == 0: reply.result = True
        return reply

    def deactivateAllQueues(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage = dallq.deactivateAllQueues()
        if reply.retCode == 0: reply.result = True
        return reply

    def enqueueLocal(self, request_iterator, context):
        phaseNo = 0
        for request in request_iterator:
            phaseNo = phaseNo + 1
            print("***************************** phaseno = %d" % phaseNo)
            print(context.peer_identities())
            errCode, errMess = self.RuleChecking(request, context)
            if errCode != 0:
                yield queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)
                break

            else:
                reply = queue_api_pb2.QAPIReply()
                reply.retCode, reply.errorMessage, reply.hdfs_file_uri = enq.enqueueLocal(request, phaseNo)
                if reply.retCode == 0: reply.result = True
                yield reply

    def distributeLocal(self, request_iterator, context):
        phaseNo = 0
        for request in request_iterator:
            phaseNo = phaseNo + 1
            print("***************************** phaseno = %d" % phaseNo)
            print(context.peer_identities())
            errCode, errMess = self.RuleChecking(request, context)
            if errCode != 0:
                yield queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)
                break

            else:
                reply = queue_api_pb2.QAPIReply()
                reply.retCode, reply.errorMessage, reply.hdfs_file_uri = distloc.distributeLocal(request, phaseNo)
                if reply.retCode == 0: reply.result = True
                yield reply

    #  top.queue_pull API functions #
    # --------------------------#

    def lander_queue_size(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage, reply.qSize = lan_q_size.lander_queue_size(request.queue_id,
                                                                                      request.caller_id)
        if reply.retCode == 0: reply.result = True
        return reply

    def lander_queue_files(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage, files = lan_q_files.lander_queue_files(request.queue_id, request.caller_id)
        if reply.retCode == 0:
            reply.result = True
            reply.listOfFilesInQueue.extend(files)
        return reply

    def lander_queue_file_older(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage = lan_q_file_old.lander_queue_file_older(request.queue_id,
                                                                                   request.hdfs_file_uri,
                                                                                   request.ageseconds,
                                                                                   request.caller_id)
        if reply.retCode == 0:
            reply.result = True
        return reply

    def lander_queue_reserve_top(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        (reply.retCode, reply.errorMessage, reply.hdfs_file_uri) = (
            lan_q_res_top.lander_queue_reserve_top(request.queue_id,
                                                   request.localtmp,
                                                   request.caller_id))
        if reply.retCode == 0:
            reply.result = True
        return reply

    def lander_queue_reserve_file(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage = lan_q_res_file.lander_queue_reserve_file(request.queue_id,
                                                                                     request.localtmp,
                                                                                     request.file_uri,
                                                                                     request.caller_id)
        if reply.retCode == 0:
            reply.result = True
        return reply

    def lander_queue_unreserve(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage = lan_q_unres.lander_queue_unreserve(request.queue_id,
                                                                               request.file_uri,
                                                                               request.caller_id)
        if reply.retCode == 0:
            reply.result = True
        return reply

    def lander_queue_release(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        (reply.retCode, reply.errorMessage) = \
            lan_q_rel.lander_queue_release(request.queue_id,
                                           request.file_uri,
                                           request.caller_id)
        if reply.retCode == 0:
            reply.result = True
        return reply

    def lander_queue_release_unconditional(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        (reply.retCode, reply.errorMessage) = \
            lan_q_rel_uncon.lander_queue_release_unconditional(request.queue_id,
                                                               request.file_uri,
                                                               request.caller_id)
        if reply.retCode == 0:
            reply.result = True
        return reply

    def lander_queue_wipe(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage = lan_q_wip.lander_queue_wipe(request.queue_id,
                                                                        request.caller_id)
        if reply.retCode == 0:
            reply.result = True
        return reply

    def markAsFaulty(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage = maf.markAsFaulty(request.queue_id,
                                                             request.file_uri,
                                                             request.caller_id)
        if reply.retCode == 0:
            reply.result = True
        return reply

    def markAsHealthy(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage = mah.markAsHealthy(request.queue_id,
                                                              request.file_uri,
                                                              request.caller_id)
        if reply.retCode == 0:
            reply.result = True
        return reply

    def listFilesInStateX(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage, files = lfisx.listFilesInStateX(request.queue_id,
                                                                           request.file_state,
                                                                           request.caller_id)
        if reply.retCode == 0:
            reply.result = True
            reply.listOfFilesInQueue.extend(files)
        return reply

    def lander_queue_list(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage, queues = lql.lander_queue_list(request.queue_id,
                                                                          request.data_type,
                                                                          request.caller_id)
        if reply.retCode == 0:
            reply.result = True
            reply.listOfQueues.extend(queues)

        return reply

    def lander_queue_du(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage, reply.queueSizeInBytes = lqd.lander_queue_du(request.queue_id,
                                                                                        request.caller_id)
        if reply.retCode == 0: reply.result = True
        return reply

    def lander_queue_user_du(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage, reply.userQueuesSizeInBytes = \
            lqud.lander_queue_user_du(request.caller_id)
        if reply.retCode == 0: reply.result = True
        return reply

    def lander_queue_queue2user(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage, reply.queueOwner = \
            lqq2u.lander_queue_queue2user(request.queue_id, request.caller_id)
        if reply.retCode == 0: reply.result = True
        return reply

    def lander_queue_user2queues(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage, queues = \
            lqu2q.lander_queue_user2queues(request.caller_id)

        if reply.retCode == 0:
            reply.result = True
            reply.listOfQueues.extend(queues)

        return reply

    def doErrorRecovery(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.fileCount, reply.errorMessage = der.doErrorRecovery(request.queue_id,
                                                                                 request.how_old,
                                                                                 request.max_fault_count,
                                                                                 request.caller_id,
                                                                                 request.is_window)
        if reply.retCode == 0: reply.result = True
        return reply

    # Regular expression related services
    # --------------------------------------------------------------------------------------------------------------------
    def getFileNameParts(self, request, context):
        #  No rule checking.  Anyone can call this function
        reply = queue_api_pb2.RegExReply()
        reply.retCode, reply.errorMessage, reply.event_time_in_utc, reply.sequence_number, \
        reply.additional_info, reply.queue_sequence_number_instance, \
        reply.queue_name = mfn.match_file_name(request.fileName)

        if reply.retCode == 0: reply.result = True
        return reply

    # stateful-streaming related services
    # --------------------------------------------------------------------------------------------------------------------
    def lander_queue_peek_next(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage, reply.count, reply.gap, reply.minutesSinceLastReserveNext, reply.hdfs_file_uri = \
            peek_next.lander_queue_peek_next(request.queue_id, request.caller_id)
        if reply.retCode == 0:
            reply.result = True
        return reply

    def lander_queue_reserve_next(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage = \
            res_next.lander_queue_reserve_next(request.queue_id, request.hdfs_file_uri, request.caller_id)
        if reply.retCode == 0:
            reply.result = True
        return reply

    #  top.queue_admin Windowing related functions. Currently directly called by Plumb only. #
    # ---------------------------------------------------------------------------------------#
    def startMaintainingWindowState(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.WindowReply(result=False, retCode=errCode, errorMessage=errMess)

        # Irrespective of wm being None or not, create a new object on this call
        print(request)
        global wm
        try:
            wm = windowmanager.WindowManager(request.windows)
        except ValueError as e:
            return queue_api_pb2.WindowReply(result=False, retCode=1, errorMessage=str(e))

        return queue_api_pb2.WindowReply(result=True, retCode=0,
                                         errorMessage="WindowManager object (wm) freshly created")

    def changeWindowState(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.WindowOperationResponse(result=False, retCode=errCode, errorMessage=errMess,
                                                         errorType=queue_api_pb2.ErrorType.OTHER)

        # If window state is not set before this call (via startMaintainingWindowState), return error
        if wm is None:
            return queue_api_pb2.WindowOperationResponse(result=False, retCode=1000,
                                                         errorMessage="Window state not set! wm is null.",
                                                         errorType=queue_api_pb2.ErrorType.UNINITIALIZED_STATE)

        reply = WindowOperationResponse()
        reply.result, reply.retCode, reply.errorMessage = wm.change_window_state(request.queueID, request.windowNo,
                                                                                 request.qSeqNoInst,
                                                                                 request.currentWindowState,
                                                                                 request.nextWindowState,
                                                                                 request.application_id)
        return reply

    def unreserveWindow(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.WindowOperationResponse(result=False, retCode=errCode, errorMessage=errMess,
                                                         errorType=queue_api_pb2.ErrorType.OTHER)

        # If window state is not set before this call (via startMaintainingWindowState), return error
        if wm is None:
            return queue_api_pb2.WindowOperationResponse(result=False, retCode=1000,
                                                         errorMessage="Window state not set! wm is null.",
                                                         errorType=queue_api_pb2.ErrorType.UNINITIALIZED_STATE)

        reply = WindowOperationResponse()
        reply.result, reply.retCode, reply.errorMessage = wm.unreserve_window(request.queueID, request.windowNo)
        return reply

    # ToDo: Should following function be re-enterent?  For example if one call is still going on in the following
    #       function and another one comes in, bad things can happen because there will be race conditions.
    #       Suggestion: (a) For now lets hope for the best. (b) Later protect critical sections with locks and if a new
    #                   call comes when old one is still pending, probably its better to fail the new call instead of
    #                   puttig it in a condition variable wait queue.  Plumb will retry anyway and hopefully, next time
    #                   call will go through.
    def getWindowsToSchedule(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.WindowReply(result=False, retCode=errCode, errorMessage=errMess,
                                             errorType=queue_api_pb2.ErrorType.OTHER)

        # If window state is not set before this call (via startMaintainingWindowState), return error
        if wm is None:
            return queue_api_pb2.WindowReply(result=False, retCode=1000,
                                             errorMessage="Window state not set! wm is null.",
                                             errorType=queue_api_pb2.ErrorType.UNINITIALIZED_STATE)

        reply = queue_api_pb2.WindowReply()
        # ToDo: we need to do for all windows.  And also one reply for each of them.  Currently for testing doing one.
        list_of_reply_windows = []
        cumulative_result = True
        cumulative_ret_code = 0
        cumulative_message = ""
        for window in request.windows:
            print("passed 1")
            r, c, m, lst = wm.get_windows_to_schedule(window.queueID)
            cumulative_result = cumulative_result and r
            cumulative_ret_code += c
            cumulative_message += m
            list_of_reply_windows += lst

        reply.result = cumulative_result
        reply.retCode = cumulative_result
        reply.errorMessage = cumulative_message
        reply.readyWindowCount = len(list_of_reply_windows)
        reply.replyWindows.extend(list_of_reply_windows)

        return reply

    def ls_windowed_queues(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.WindowReply(result=False, retCode=errCode, errorMessage=errMess,
                                             errorType=queue_api_pb2.ErrorType.OTHER)

        # If window state is not set before this call (via startMaintainingWindowState), return error
        if wm is None:
            return queue_api_pb2.WindowReply(result=False, retCode=1000,
                                             errorMessage="Window state not set! wm is null. Run Plumb to set state",
                                             errorType=queue_api_pb2.ErrorType.UNINITIALIZED_STATE)

        r, c, m, q_list = wm.ls_windowed_queues()
        reply = queue_api_pb2.WindowReply()
        reply.result = r
        reply.retCode = c
        reply.errorMessage = m
        reply.listOfQueues.extend(q_list)

        return reply

    def ls_all_window_statuses(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.WindowReply(result=False, retCode=errCode, errorMessage=errMess,
                                             errorType=queue_api_pb2.ErrorType.OTHER)

        # If window state is not set before this call (via startMaintainingWindowState), return error
        if wm is None:
            return queue_api_pb2.WindowReply(result=False, retCode=1000,
                                             errorMessage="Window state not set! wm is null. Run Plumb to set state",
                                             errorType=queue_api_pb2.ErrorType.UNINITIALIZED_STATE)

        r, c, m, statuses = wm.ls_all_window_statuses()
        reply = queue_api_pb2.WindowReply()
        reply.result = r
        reply.retCode = c
        reply.errorMessage = m
        reply.listOfStatuses.extend(statuses)

        return reply

    def ls_win(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.WindowReply(result=False, retCode=errCode, errorMessage=errMess,
                                             errorType=queue_api_pb2.ErrorType.OTHER)

        # If window state is not set before this call (via startMaintainingWindowState), return error
        if wm is None:
            return queue_api_pb2.WindowReply(result=False, retCode=1000,
                                             errorMessage="Window state not set! wm is null. Run Plumb to set state",
                                             errorType=queue_api_pb2.ErrorType.UNINITIALIZED_STATE)

        r, c, m, statuses = wm.ls_win(request.queue_id)
        reply = queue_api_pb2.WindowReply()
        reply.result = r
        reply.retCode = c
        reply.errorMessage = m
        reply.listOfStatuses.extend(statuses)

        return reply

    def ls_win_in_status(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.WindowReply(result=False, retCode=errCode, errorMessage=errMess,
                                             errorType=queue_api_pb2.ErrorType.OTHER)

        # If window state is not set before this call (via startMaintainingWindowState), return error
        if wm is None:
            return queue_api_pb2.WindowReply(result=False, retCode=1000,
                                             errorMessage="Window state not set! wm is null. Run Plumb to set state",
                                             errorType=queue_api_pb2.ErrorType.UNINITIALIZED_STATE)

        r, c, m, windows = wm.ls_win_in_status(request.queue_id, request.status)
        reply = queue_api_pb2.WindowReply()
        reply.result = r
        reply.retCode = c
        reply.errorMessage = m
        reply.listOfWindows.extend(windows)

        return reply

    def set_win_status(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.WindowReply(result=False, retCode=errCode, errorMessage=errMess,
                                             errorType=queue_api_pb2.ErrorType.OTHER)

        # If window state is not set before this call (via startMaintainingWindowState), return error
        if wm is None:
            return queue_api_pb2.WindowReply(result=False, retCode=1000,
                                             errorMessage="Window state not set! wm is null. Run Plumb to set state",
                                             errorType=queue_api_pb2.ErrorType.UNINITIALIZED_STATE)

        r, c, m = wm.set_win_status(request.queue_id, request.window_no, request.status)
        reply = queue_api_pb2.WindowReply()
        reply.result = r
        reply.retCode = c
        reply.errorMessage = m
        return reply

    def ls_f_in_win(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.WindowReply(result=False, retCode=errCode, errorMessage=errMess,
                                             errorType=queue_api_pb2.ErrorType.OTHER)

        # If window state is not set before this call (via startMaintainingWindowState), return error
        if wm is None:
            return queue_api_pb2.WindowReply(result=False, retCode=1000,
                                             errorMessage="Window state not set! wm is null. Run Plumb to set state",
                                             errorType=queue_api_pb2.ErrorType.UNINITIALIZED_STATE)

        r, c, m, files = wm.ls_f_in_win(request.queue_id, request.window_no)
        reply = queue_api_pb2.WindowReply()
        reply.result = r
        reply.retCode = c
        reply.errorMessage = m
        reply.listOfFiles.extend(files)
        return reply

    def mv_f_in_win(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.WindowReply(result=False, retCode=errCode, errorMessage=errMess,
                                             errorType=queue_api_pb2.ErrorType.OTHER)

        # If window state is not set before this call (via startMaintainingWindowState), return error
        if wm is None:
            return queue_api_pb2.WindowReply(result=False, retCode=1000,
                                             errorMessage="Window state not set! wm is null. Run Plumb to set state",
                                             errorType=queue_api_pb2.ErrorType.UNINITIALIZED_STATE)

        r, c, m = wm.mv_f_in_win(request.queue_id, request.window_no, request.file_uri)
        reply = queue_api_pb2.WindowReply()
        reply.result = r
        reply.retCode = c
        reply.errorMessage = m
        return reply

    def ls_missing_files_in_a_window(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.WindowReply(result=False, retCode=errCode, errorMessage=errMess,
                                             errorType=queue_api_pb2.ErrorType.OTHER)

        # If window state is not set before this call (via startMaintainingWindowState), return error
        if wm is None:
            return queue_api_pb2.WindowReply(result=False, retCode=1000,
                                             errorMessage="Window state not set! wm is null. Run Plumb to set state",
                                             errorType=queue_api_pb2.ErrorType.UNINITIALIZED_STATE)

        r, c, m, files = wm.ls_missing_files_in_a_window(request.queue_id)
        reply = queue_api_pb2.WindowReply()
        reply.result = r
        reply.retCode = c
        reply.errorMessage = m
        reply.listOfFiles.extend(files)
        return reply

    def ls_site_info_in_a_window(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.WindowReply(result=False, retCode=errCode, errorMessage=errMess,
                                             errorType=queue_api_pb2.ErrorType.OTHER)

        # If window state is not set before this call (via startMaintainingWindowState), return error
        if wm is None:
            return queue_api_pb2.WindowReply(result=False, retCode=1000,
                                             errorMessage="Window state not set! wm is null. Run Plumb to set state",
                                             errorType=queue_api_pb2.ErrorType.UNINITIALIZED_STATE)

        r, c, m, files = wm.ls_site_info_in_a_window(request.queue_id)
        reply = queue_api_pb2.WindowReply()
        reply.result = r
        reply.retCode = c
        reply.errorMessage = m
        reply.siteInfoList.extend(files)
        return reply

    def lander_queue_list_keys(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage, keys = lan_q_ls_keys.lander_queue_list_keys(request.queue_id,
                                                                                       request.caller_id)
        if reply.retCode == 0:
            reply.result = True
            reply.listOfKeysInKeyVal.extend(keys)
        return reply

    def get_value(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage, reply.value = gv.get_value(request.queue_id, request.caller_id, request.key)
        if reply.retCode == 0: reply.result = True
        return reply

    def put_value(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage = pv.put_value(request.queue_id, request.caller_id, request.key,
                                                         request.value)
        if reply.retCode == 0:
            reply.result = True
        return reply

    def del_value(self, request, context):
        print(context.peer_identities())
        errCode, errMess = self.RuleChecking(request, context)
        if errCode != 0:
            return queue_api_pb2.QAPIReply(result=False, retCode=errCode, errorMessage=errMess)

        reply = queue_api_pb2.QAPIReply()
        reply.retCode, reply.errorMessage = dv.del_value(request.queue_id, request.caller_id, request.key,
                                                         request.value)
        if reply.retCode == 0: reply.result = True
        return reply


def serve():
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    queue_api_pb2_grpc.add_LanderQueueAPIImplementationServicer_to_server(LanderQueueAPIImplementation(), server)

    server_key_pem = "%s/server.key.pem" % KeyPath
    certChain = "%s/server_chain" % KeyPath
    rootCA = "%s/roots.pem" % KeyPath

    credentials = grpc.ssl_server_credentials(((open(server_key_pem, 'rb').read(),
                                                open(certChain, 'rb').read()),),
                                              open(rootCA, 'rb').read(),
                                              True)
    # server.add_insecure_port(':::50051')
    server.add_secure_port('[::]:50051', credentials)
    server.start()

    ##global wm
    ##wm  = windowmanager.WindowManager()

    try:
        while True:
            time.sleep(_ONE_DAY_IN_SECONDS)
    except KeyboardInterrupt:
        server.stop(0)


if __name__ == '__main__':
    serve()
