#!/usr/bin/env python3

import pyhdfs
import getpass
import sys

from top.conf import sysConfInfo



def e(ret):
    exit(ret)



# for reference code of pyhdfs is at: https://github.com/jingw/pyhdfs/blob/master/pyhdfs.py
# documentation at: https://pyhdfs.readthedocs.io/en/latest/pyhdfs.html#pyhdfs.HdfsClient.open
# and some high level details here: https://pypi.org/project/PyHDFS/


_DEBUG_THIS = True
def _debug(s):
  if _DEBUG_THIS:
    sys.stderr.write("DEBUG: " + str(s) + "\n")


def getHDFSReference():
    whoami = getpass.getuser()
    nameNode = "%s:%s"  %(sysConfInfo.HDFS_NAMENODE, sysConfInfo.HDFS_PORT)
    return pyhdfs.HdfsClient(hosts=nameNode, user_name=whoami)
    


def makeNewHDFSDir(hdfsURI):
    result = False
    try:
        fs = getHDFSReference()
        result = fs.mkdirs(hdfsURI)
    except Exception as e:
        _debug(e)
    return 0 if result else 1



def makeHDFSDirIfDoesNotExist(hdfsURI):
    return makeNewHDFSDir(hdfsURI)



def doesHDFSDirExists(hdfsURI):
    fileType = ""
    try:
        fs = getHDFSReference()
        fileType=fs.get_file_status(hdfsURI).type
    except Exception as e:
        _debug(e)
    return 0 if fileType=='DIRECTORY' else 2


def setFileOrDirRights(hdfsURI, rights):
    retCode = 0
    try:
        fs = getHDFSReference()
        fs.set_permission(hdfsURI, permission=rights)
    except Exception as e:
        retCode = 3
        _debug(e)
    return retCode


def setFileOrDirGroup(hdfsURI, grp):
    retCode = 0
    try:
        fs = getHDFSReference()
        fs.set_owner(hdfsURI, group=grp)
    except Exception as e:
        retCode = 4
        _debug(e)
    return retCode


def doesHDFSFileExists(hdfsURI):
    retCode = 0
    try:
        fs = getHDFSReference()
        fileType=fs.get_file_status(hdfsURI).type
        retCode = 0 if fileType=='FILE' else 5
    except Exception as e:
        retCode = 6
        _debug(e)
    return retCode



def copyFileFromLocal(localURI, hdfsURI, shouldOverWrite):
    retCode = 0
    try:
        fs = getHDFSReference() 
        if shouldOverWrite == True:
            fs.copy_from_local(localURI, hdfsURI, overwrite=True)
        else:
            fs.copy_from_local(localURI, hdfsURI)
    except Exception as e:
        retCode = 7
        _debug(e)
    return retCode







def copyToLocal(localURI, hdfsURI):
    retCode = 0
    try:
        fs = getHDFSReference() 
        fs.copy_to_local(hdfsURI, localURI)
    except Exception as e:
        retCode = 8
        _debug(e)
    return retCode





def delHDFSFile(hdfsURI):
    result = False
    try:
        fs = getHDFSReference()
        result = fs.delete(hdfsURI)
    except Exception as e:
        _debug(e)
    return 0 if result==True else 9




def mvHDFSFile(srcHdfsURI, dstHdfsURI):
    result = False
    try:
        fs = getHDFSReference()
        result = fs.rename(srcHdfsURI, dstHdfsURI)
    except Exception as e:
        _debug(e)
    return 0 if result==True else 10





def getHDFSFileModTime(hdfsURI):
    retCode = 0
    modTime = 0
    try:
        fs = getHDFSReference()
        modTime=fs.get_file_status(hdfsURI).modificationTime
    except Exception as e:
        retCode = 11
        _debug(e)
    return retCode, modTime




def getHDFSFileSizeInBytes(hdfsURI):
    retCode = 0
    noOfBytes = 0
    try:
        fs = getHDFSReference()
        noOfBytes = fs.get_content_summary(hdfsURI).length
    except Exception as e:
        retCode = 12
        _debug(e)
    return retCode, noOfBytes



# ToFix: Following is providing MD5-of-262144MD5-of-512CRC32C while we
# need CRC32C of all data.  Currently it is not clear how to get that
# using this pyHDFS API.  That is possible to do with hdfs command as follows:
# hdfs -Ddfs.checksum.combine.mode=COMPOSITE_CRC -checksum /user/aqadeer/tmp/x.pcap.sz
#
def getHDFSFileCRC32c(hdfsURI):
    retCode = 0
    crc32c = ""
    try:
        fs = getHDFSReference()
        crc32c = fs.get_file_checksum(hdfsURI, algorithm="COMPOSITE_CRC").bytes
        print (fs.get_file_checksum(hdfsURI).algorithm)
        print (crc32c)
    except Exception as e:
        retCode = 13
        _debug(e)
    return retCode, crc32c



# for testing 
#if __name__ == '__main__':
    #print makeNewHDFSDir("/user/aqadeer/unittest")
    #print makeHDFSDirIfDoesNotExist("/user/aqadeer/unittest")
    #print doesHDFSDirExists("/user/aqadeer/unittest");
    #print doesHDFSDirExists("/user/aqadeer/unittest2");
    #print setFileOrDirRights("/user/aqadeer/unittest", "900")
    #print setFileOrDirRights("/user/aqadeer/unittest", "777")
    #print setFileOrDirGroup("/user/aqadeer/unittest2", "cp-raw")
    #print setFileOrDirGroup("/user/aqadeer/unittest", "cp-raw")
    #print setFileOrDirGroup("/user/aqadeer/unittest2", "cp-raww")
    #print copyFileFromLocal("/tmp/test.txt", "/user/aqadeer/unittest", False);
    #print copyFileFromLocal("/tmp/test.txt", "/user/aqadeer/unittest", False);
    #print copyFileFromLocal("/tmp/test.txt", "/user/aqadeer/unittest", True);
    #print copyFileFromLocal("/tmp/test.txtt", "/user/aqadeer/unittest", False);
    #print copyFileFromLocal("/tmp/test.txt", "/user/aqadeer/unittest2", False);
    #print doesHDFSFileExists("/user/aqadeer/unittest2");
    #print doesHDFSFileExists("/user/aqadeer/unittest");
    #print doesHDFSFileExists("/user/aqadeer/unittest23");
    #print copyToLocal("/tmp/test2.txt", "/user/aqadeer/unittest2");
    # Tofix: How to pass stdout as second parameter
    #print copyToLocal("/user/aqadeer/tmp/x.pcap.sz", sys.stdout)
    #print copyToLocal("/tmp/test2.txt", "/user/aqadeer/unittest23");
    #print copyToLocal("/tmp/test.txt", "/user/aqadeer/unittest2");
    #print getHDFSFileSizeInBytes("/user/aqadeer/unittest2");
    #print getHDFSFileSizeInBytes("/user/aqadeer/unittest");
    #print getHDFSFileSizeInBytes("/user/aqadeer/unittest23");
    #print getHDFSFileModTime("/user/aqadeer/unittest2")
    #print getHDFSFileModTime("/user/aqadeer/unittest")
    #print getHDFSFileModTime("/user/aqadeer/unittest23")
    #print mvHDFSFile("/user/aqadeer/unittest2", "/user/aqadeer/unittest/test.txt")
    #print mvHDFSFile("/user/aqadeer/unittest23", "/user/aqadeer/unittest/test.txt")
    #print delHDFSFile("/user/aqadeer/a")
    #print delHDFSFile("/user/aqadeer/unittest/test.txt")
    #print delHDFSFile("/user/aqadeer/unittest")
    #print delHDFSFile("/user/aqadeer/unittest23")
    #print getHDFSFileCRC32c("/user/aqadeer/tmp/x.pcap.sz")

