#!/usr/bin/env python3

import re
import datetime

debug = False


def match_file_name(file_name, queue_name_from_yaml=None):
    # The r in the following implies, the string is a raw string, which means all escape codes will be ignored.
    # Be careful with group numbering.  We have 8 groups in the following.
    match_string = r'(\d{8})-(\d{6})-(\d+)((-([^.]+))?)\.([^.]+)\.(.+)'
    matcher = re.match(match_string, file_name)

    if debug:
        print("file_name = %s  and queue_name_from_yaml = %s" % (file_name, queue_name_from_yaml))
    min_length = 1
    max_length = 767
    queue_name_max_length = 255
    date = "Uninitialized"
    sequence_no = "Uninitialized"
    optional_info = ""  # This is optional part.  So an empty string implies no optional info to bash like programs.
    sequence_no_instance = "Uninitialized"
    queue_name = "Uninitialized"

    if matcher:
        if debug:
            print(matcher.groups())
        date = matcher.group(1) + "-" + matcher.group(2)
        try:
            datetime.datetime.strptime(date, "%Y%m%d-%H%M%S")
        except ValueError as err:
            # print(err)
            return 1, "Date not valid. Required format (YYYYmmdd-HHMMSS). %s" % err, \
                   date, sequence_no, optional_info, sequence_no_instance, queue_name

        sequence_no = matcher.group(3)
        if len(sequence_no) < min_length or len(sequence_no) > max_length:
            return 3, "Sequence no length should be between [%s, %s] digits" % (min_length, max_length), \
                   date, sequence_no, optional_info, sequence_no_instance, queue_name

        optional_info = matcher.group(6)
        if (optional_info is not None) and (len(optional_info) > max_length):
            return 4, "Optional meta data info should not exceed %s characters." % max_length, \
                   date, sequence_no, optional_info, sequence_no_instance, queue_name

        if optional_info is None:
            optional_info = ""

        sequence_no_instance = matcher.group(7)
        if len(sequence_no_instance) < min_length or len(sequence_no_instance) > max_length:
            return 7, "Sequence number instance should be between [%s,%s] characters long" % (min_length, max_length), \
                   date, sequence_no, optional_info, sequence_no_instance, queue_name

        queue_name = matcher.group(8)
        # Note: We don't need special handing for queue None because this queue should never occur as input queue.
        if len(queue_name) < min_length or len(queue_name) > queue_name_max_length \
                or (queue_name_from_yaml is not None and queue_name != queue_name_from_yaml):
            return 8, "Queue name should match with YAML description and queue length should be between [%s,%s] chars" \
                   % (min_length, queue_name_max_length), \
                   date, sequence_no, optional_info, sequence_no_instance, queue_name

        return 0, "OK", date, sequence_no, optional_info, sequence_no_instance, queue_name
    else:
        return 9, "Name parsing failed. Not in required format %s" % match_string, \
               date, sequence_no, optional_info, sequence_no_instance, queue_name


def get_event_time_format():
    return "%Y%m%d-%H%i%s"


def get_event_time_in_utc(file_name):
    ret_code, err_message, date, sequence_no, optional_info, sequence_no_instance, queue_name = \
        match_file_name(file_name)
    # Following printing is for bash like set-up
    print(ret_code)
    print(err_message)
    print(date)
    return ret_code, err_message, date


def get_ymt(file_name):
    ret_code, err_message, date, sequence_no, optional_info, sequence_no_instance, queue_name = \
        match_file_name(file_name)
    # Following printing is for bash like set-up
    print(ret_code)
    print(err_message)
    if ret_code == 0: print(date[0:8])
    return ret_code, err_message, date[0:8]


def get_sequence_number(file_name):
    ret_code, err_message, date, sequence_no, optional_info, sequence_no_instance, queue_name = \
        match_file_name(file_name)
    # Following printing is for bash like set-up
    print(ret_code)
    print(err_message)
    print(sequence_no)
    return ret_code, err_message, sequence_no


def get_additional_info(file_name):
    ret_code, err_message, date, sequence_no, optional_info, sequence_no_instance, queue_name = \
        match_file_name(file_name)
    # Following printing is for bash like set-up
    print(ret_code)
    print(err_message)
    print(optional_info)
    return ret_code, err_message, optional_info


def get_queue_sequence_number_instance(file_name):
    ret_code, err_message, date, sequence_no, optional_info, sequence_no_instance, queue_name = \
        match_file_name(file_name)
    # Following printing is for bash like set-up
    print(ret_code)
    print(err_message)
    print(sequence_no_instance)
    return ret_code, err_message, sequence_no_instance


def get_queue_name(file_name):
    ret_code, err_message, date, sequence_no, optional_info, sequence_no_instance, queue_name = \
        match_file_name(file_name)
    # Following printing is for bash like set-up
    print(ret_code)
    print(err_message)
    print(queue_name)
    return ret_code, err_message, queue_name


# # for testing
# if __name__ == '__main__':
#     queue_name_f_yaml = "pcap.xz"
#     file_name1 = "20200413-212204-01779440.lax.pcap.xz"
#     file_name2 = "20200413-212204-01779440.lax.dummyinfo.pcap.xz"
#     file_name3 = "20200413-212204-25.laxprocessingsite.pcap.xz"
#
#     file_name4 = "2000413-212204-25.laxprocessingsite.pcap.xz"
#     file_name5 = "20200413-262204-25.laxprocessingsite.pcap.xz"
#     file_name6 = "20200413-212204-253333333333333333333.pcap.xz"
#
#     file_name7 = "20200504-221112-00000001-us_east_1f-1588617573-ac1f337b-20011878040000000000008009a092.aws.pcap.sz"
#
#     # retCode, retMessage, date, blockNo, site, tail = matchFileName(file_name)
#     print(match_file_name(file_name1, queue_name_f_yaml))
#     print(match_file_name(file_name2, queue_name_f_yaml))
#     print(match_file_name(file_name3, queue_name_f_yaml))
#     print(match_file_name(file_name4, queue_name_f_yaml))
#     print(match_file_name(file_name5, queue_name_f_yaml))
#     print(match_file_name(file_name6, queue_name_f_yaml))
#
#     print(match_file_name(file_name1, "pcap.sz"))
#     print(match_file_name("20200413-214831-00335202.mia.pcap.xz", "pcap.sz"))
#
#     print(match_file_name(file_name7, "pcap.sz"))
#
#     print(match_file_name("20200413-2122004-01779440.lax.pcap.xz", "pcap.xz"))
#     print(match_file_name("20200413-2122004-0177a9440.lax.pcap.xz", "pcap.xz"))
#     print(match_file_name("20200413-2122004-01779440.some.info.lax.pcap.xz", "pcap.xz"))
#
#     get_event_time_in_utc("20200504-221112-00000001-us_east_1f-1588617573-ac1f337b-20011800000008009a092.aws.pcap.sz")
#     get_sequence_number("20200504-221112-00000001-us_east_1f-1588617573-ac1f337b-20011800000008009a092.aws.pcap.sz")
#     get_additional_info("20200504-221112-00000001-us_east_1f-1588617573-ac1f337b-20011800000008009a092.aws.pcap.sz")
#     get_queue_sequence_number_instance("20200504-221112-00000001-us_east_1f-15886b-20011800000008009a092.aws.pcap.sz")
#     get_queue_name("20200504-221112-00000001-us_east_1f-1588617573-ac1f337b-20011800000008009a092.aws.pcap.sz")
#
#     get_additional_info("20200504-221112-00000001.aws.pcap.sz")
#     get_additional_info("20200504-221112-0a0000001.aws.pcap.sz")
